""" Module for variant peptide pool (unique) """
from __future__ import annotations
from typing import Set, IO, Dict, List, Tuple, TYPE_CHECKING
from pathlib import Path
from Bio import SeqUtils, SeqIO
from Bio.Seq import Seq
from Bio.SeqIO import FastaIO
from moPepGen.aa.AminoAcidSeqRecord import AminoAcidSeqRecord
from moPepGen import get_equivalent, VARIANT_PEPTIDE_SOURCE_DELIMITER
from .VariantPeptideLabel import VariantPeptideInfo


if TYPE_CHECKING:
    from moPepGen import params

class VariantPeptidePool():
    """ Varaint Peptide Pool """
    def __init__(self, peptides:Set[AminoAcidSeqRecord]=None):
        """ Constructor """
        self.peptides = peptides or set()
        self.peptide_delimeter = VARIANT_PEPTIDE_SOURCE_DELIMITER

    def add_peptide(self, peptide:AminoAcidSeqRecord,
            canonical_peptides:Set[str], cleavage_params:params.CleavageParams=None,
            skip_checking:bool=False) -> bool:
        """ Add a peptide to the pool if it does not already exist. Otherwise,
        the label is appended to the existing same peptide.

        Args:
            peptide (AminoAcidSeqRecord): The amino acid sequence to be added.
            canonical_peptides (Set[str]): Canonical peptides.
            cleavage_params (CleavageParams): cleavage parameters.
            skip_checking (bool): whether to skip checking cleavage parameters.
        """
        if not skip_checking:
            min_mw = cleavage_params.min_mw
            min_length = cleavage_params.min_length
            max_length = cleavage_params.max_length
            if SeqUtils.molecular_weight(peptide.seq, 'protein') < min_mw:
                return False
            if len(peptide.seq) < min_length or len(peptide.seq) > max_length:
                return False
            if str(peptide.seq) in canonical_peptides:
                return False
        same_peptide = get_equivalent(self.peptides, peptide)
        if same_peptide:
            same_peptide:Seq
            new_label = peptide.description
            same_peptide.description += (self.peptide_delimeter + new_label)
            same_peptide.id = same_peptide.description
            same_peptide.name = same_peptide.description
        else:
            self.peptides.add(peptide)
        return True

    def remove_redundant_headers(self):
        """ remove redundant fasta header entries """
        for peptide in self.peptides:
            entries = {}
            for entry in peptide.description.split(self.peptide_delimeter):
                unversioned = entry.rsplit('|', 1)[0]
                if unversioned in entries:
                    continue
                entries[unversioned] = entry
            header = self.peptide_delimeter.join(entries.values())
            peptide.description = header
            peptide.id = header
            peptide.name = header

    def write(self, path:Path):
        """ Write the variant peptide pool to FASTA file.

        Args:
            path (Path): Output FASTA file path.
        """
        with open(path, 'w') as handle:
            record2title = lambda x: x.description
            writer = FastaIO.FastaWriter(handle, record2title=record2title)
            for record in self.peptides:
                writer.write_record(record)

    @classmethod
    def load(cls, handle:IO) -> VariantPeptidePool:
        """ Load variant peptide pool from file """
        pool = cls()
        for seq in SeqIO.parse(handle, 'fasta'):
            seq.__class__ = AminoAcidSeqRecord
            seq.id = seq.description
            seq.name = seq.description
            pool.peptides.add(seq)
        return pool

    def filter(self, exprs:Dict[str,int]=None, cutoff:float=None,
            coding_transcripts:List[str]=None, keep_all_noncoding:bool=False,
            keep_all_coding:bool=False, enzyme:str='trypsin',
            miscleavage_range:Tuple[int,int]=(None, None,),
            denylist:Set[Seq]=None, keep_canonical:bool=False) -> VariantPeptidePool:
        """ Filter variant peptides according to gene expression. """
        label_delimiter = VARIANT_PEPTIDE_SOURCE_DELIMITER
        filtered_pool = VariantPeptidePool()
        for peptide in self.peptides:
            # Filter by miscleavages
            if any(x is not None for x in miscleavage_range):
                exception = 'trypsin_exception' if enzyme == 'trypsin' else None
                misc = peptide.find_all_enzymatic_cleave_sites(enzyme, exception)
                if miscleavage_range[0] is not None and len(misc) < miscleavage_range[0]:
                    continue
                if miscleavage_range[1] is not None and len(misc) > miscleavage_range[1]:
                    continue

            peptide_entries = VariantPeptideInfo.from_variant_peptide_minimal(peptide)

            is_in_denylist = denylist is not None and peptide.seq in denylist
            keep = []
            for entry in peptide_entries:
                all_noncoding = not any(x in coding_transcripts
                    for x in entry.get_transcript_ids())
                all_coding = all(x in coding_transcripts
                    for x in entry.get_transcript_ids())

                is_canonical = ((not entry.is_circ_rna()) and \
                    entry.get_transcript_ids()[0] in coding_transcripts)

                if is_in_denylist and (not (keep_canonical and is_canonical)):
                    should_keep = False

                elif keep_all_noncoding and all_noncoding:
                    should_keep = True

                elif keep_all_coding and all_coding:
                    should_keep = True

                else:
                    if exprs is not None:
                        tx_ids = entry.get_transcript_ids()
                        should_keep = entry.is_fusion() or entry.is_circ_rna() or\
                            entry.is_splice_altering() or \
                            all(exprs[tx] >= cutoff for tx in tx_ids)
                    else:
                        should_keep = True
                if should_keep:
                    keep.append(entry)
            if keep:
                label = label_delimiter.join([str(x) for x in keep])
                peptide.description = label
                filtered_pool.peptides.add(peptide)
        return filtered_pool



""" Add fuzz test log """
import argparse
import datetime
import subprocess as sp
from typing import IO
from pathlib import Path
from moPepGen.cli.common import print_help_if_missing_args, add_args_debug_level
from moPepGen.util.fuzz_test import FuzzRecord, FuzzRecordStatus


FUZZ_TEST_LOG_HISTORY_CSV = Path(__file__).parent.parent.parent/'docs/files/fuzz_test_history.tsv'

def get_std(m:float, ss:float, n:int):
    """ Calculate standard deviation.

    Args:
        - m (float): mean
        - ss (float): sum of square
        - n (int): size
    """
    return (ss/n - m**2) ** (1/2)

# pylint: disable=W0212
def parse_args(subparsers:argparse._SubParsersAction):
    """ parse args """
    p:argparse.ArgumentParser = subparsers.add_parser(
        name='addFuzzTestLog',
        help='Add fuzz test log.'
    )
    p.add_argument(
        '-l', '--log-files',
        type=Path,
        help='Fuzz test log files',
        nargs='+',
        metavar='<files>',
        required=True
    )
    p.add_argument(
        '-c', '--commit-id',
        type=str,
        help='The Git commit ID that the fuzz tests used.',
        metavar='<value>',
        default=None
    )
    p.set_defaults(func=main)
    add_args_debug_level(p)
    print_help_if_missing_args(p)
    return p

class FuzzTestLogSummary:
    """ Fuzz test log summary """
    _commit = None
    _version = None

    def __init__(self, commit:str=None, version:str=None,
            n_match:int=0, n_mismatch:int=0, n_fail:int=0,
            avg_time_call_variant:datetime.timedelta=None,
            ss_time_call_variant:int=0,
            avg_time_brute_force:datetime.timedelta=None,
            ss_time_brute_force:int=0):
        """ constructor """
        self.commit = commit or self.get_commit()
        self.version = version or self.get_version()
        self.n_match = n_match
        self.n_mismatch = n_mismatch
        self.n_fail = n_fail
        self.avg_time_call_variant = avg_time_call_variant or datetime.timedelta()
        self.ss_time_call_variant = ss_time_call_variant
        self.avg_time_brute_force = avg_time_brute_force or datetime.timedelta()
        self.ss_time_brute_force = ss_time_brute_force

    @classmethod
    def get_commit(cls) -> str:
        """ Get the latest commit ID """
        if cls._commit:
            return cls._commit
        out = sp.check_output(['git', 'rev-parse', '--short', 'HEAD'])
        commit = out.decode('utf-8').rstrip()
        cls._commit = commit
        return commit

    @classmethod
    def get_version(cls) -> str:
        """ Get the latest version """
        if cls._version:
            return cls._version
        out = sp.check_output(['git', 'describe', '--tags', '--abbrev=0'])
        version = out.decode('utf-8').rstrip()
        cls._version = version
        return version

    def add_record(self, record:FuzzRecord):
        """ Add record """
        n_before = self.n_match + self.n_mismatch
        if record.status == FuzzRecordStatus.succeeded:
            self.n_match += 1
        elif record.status == FuzzRecordStatus.failed:
            self.n_mismatch += 1
            return
        else:
            self.n_fail += 1
            return

        dt_call_variant = (record.call_variant_end - record.call_variant_start)
        self.avg_time_call_variant = (
            self.avg_time_call_variant * n_before + dt_call_variant
        ) / (n_before + 1)
        self.ss_time_call_variant += dt_call_variant.total_seconds() ** 2

        dt_brute_force = (record.brute_force_end - record.brute_force_start)
        self.avg_time_brute_force = (
            self.avg_time_brute_force * n_before + dt_brute_force
        ) / (n_before + 1)
        self.ss_time_brute_force += dt_brute_force.total_seconds() ** 2


class FuzzTestLogSummaryCategorized:
    """ Fuzz test log summary separate by categories """
    def __init__(self, snv:FuzzTestLogSummary=None, indel:FuzzTestLogSummary=None,
            comprehensive:FuzzTestLogSummary=None, submit_date:datetime.date=None,
            commit:str=None):
        """ Constructor """
        self.snv = snv or FuzzTestLogSummary(commit=commit)
        self.indel = indel or FuzzTestLogSummary(commit=commit)
        self.comprehensive = comprehensive or FuzzTestLogSummary(commit=commit)
        self.submit_date = submit_date

    def summarize(self, handle:IO):
        """ Summarize fuzz run summary from the log file. """
        for record in FuzzRecord.parse(handle):
            if record.n_alt_splice == 0 \
                    and record.n_circ_rna == 0 \
                    and record.n_fusion == 0:
                if record.n_indel == 0:
                    self.snv.add_record(record)
                else:
                    self.indel.add_record(record)
            else:
                self.comprehensive.add_record(record)

            submitted = record.submitted.date()
            if self.submit_date is None or self.submit_date > submitted:
                self.submit_date = submitted

    @staticmethod
    def tsv_header() -> str:
        """ get TSV header """
        return '\t'.join([
            'version', 'commit', 'submit_date', 'variant_type',
            'n_match', 'n_mismatch', 'n_fail',
            'avg_time_call_variant', 'std_sec_call_variant',
            'avg_time_brute_force', 'std_sec_brute_force'
        ])

    def append_to_history(self, handle:IO):
        """ append summary to handle """
        if handle.tell() == 0:
            handle.write(self.tsv_header() + '\n')
        summaries = {
            'snv': self.snv,
            'indel': self.indel,
            'comprehensive': self.comprehensive
        }
        for key, summary in summaries.items():
            std_call_variant = get_std(
                m=summary.avg_time_call_variant.total_seconds(),
                ss=summary.ss_time_call_variant,
                n=summary.n_match
            )
            std_brute_force = get_std(
                m=summary.avg_time_brute_force.total_seconds(),
                ss=summary.ss_time_brute_force,
                n=summary.n_match
            )
            fields = [
                summary.version,
                summary.commit,
                str(self.submit_date),
                key,
                str(summary.n_match),
                str(summary.n_mismatch),
                str(summary.n_fail),
                str(summary.avg_time_call_variant),
                str(std_call_variant),
                str(summary.avg_time_brute_force),
                str(std_brute_force)
            ]
            handle.write('\t'.join(fields) + '\n')


def main(args:argparse.Namespace):
    """ Add fuzz test log """
    summary = FuzzTestLogSummaryCategorized(commit=args.commit_id)

    for file in args.log_files:
        with open(file, 'rt') as handle:
            summary.summarize(handle)

    with open(FUZZ_TEST_LOG_HISTORY_CSV, 'at') as handle:
        summary.append_to_history(handle)



""" A brute forth algorithm for calling variant peptides from a GVF file. """
import copy
import sys
import argparse
import math
from typing import Iterable, List, Dict, Set, Tuple
from pathlib import Path
from itertools import combinations
from Bio import SeqUtils
from Bio.Seq import Seq
from moPepGen import gtf, seqvar, aa, dna, params, constant
from moPepGen.SeqFeature import FeatureLocation, SeqFeature
from moPepGen.cli import common as cli_common
from moPepGen.seqvar.VariantRecord import VariantRecord
from moPepGen.seqvar.VariantRecordPool import VariantRecordPool
from moPepGen.seqvar.VariantRecordWithCoordinate import VariantRecordWithCoordinate
from moPepGen.util.common import load_references


# pylint: disable=W0212
def parse_args(subparsers:argparse._SubParsersAction):
    """ parse command line arguments """
    parser:argparse.ArgumentParser = subparsers.add_parser(
        name='bruteForce',
        help='Call variant peptide with the brute force algorithm.'
    )
    parser.add_argument(
        '-i', '--input-gvf',
        type=Path,
        help='GVF file',
        nargs="+"
    )
    parser.add_argument(
        '-r', '--reference-dir',
        type=Path,
        help='Reference directory. Must contain genome.fa, annotation.gtf, and'
        ' proteome.fasta. The directory should be generated by'
        ' the downsampleReference command'
    )
    parser.add_argument(
        '-f', '--force',
        action='store_true',
        help='If not set, the program stops when there are more than 10'
        ' variants. When this flag is set, the program runs anyway. Noted that '
        ' the runtime is going to increase quickly after 10 variants.',
        default=False
    )
    parser.add_argument(
        '--variant-ids',
        type=str,
        help='List of variant labels.',
        nargs='*'
    )
    parser.add_argument(
        '--max-adjacent-as-mnv',
        type=int,
        help='Max number of adjacent SNPs or INDELs to be merged as a MNV.',
        default=2
    )
    parser.add_argument(
        '--selenocysteine-termination',
        action='store_true',
        help='Include peptides of selenoprotiens that the UGA is treated as '
        'termination instead of Sec.'
    )
    parser.add_argument(
        '--w2f-reassignment',
        action='store_true',
        help='Include peptides with W > F (Tryptophan to Phenylalanine) '
        'reassignment.'
    )
    cli_common.add_args_cleavage(parser)
    cli_common.add_args_debug_level(parser)

    parser.set_defaults(func=main)
    cli_common.print_help_if_missing_args(parser)
    return parser

def _parse_exclusion(exclusion) -> Tuple:
    """ Parse exclusion to values """
    start, ref, alt = exclusion.split('-')
    return int(start), ref, alt

def parse_variant_exclusion(exclusions:List[str]) -> Dict[Tuple,List[Tuple]]:
    """ Parse exclusion variants into """
    groups = {}
    for exclusion in exclusions:
        variant, targets = exclusion.split(':')
        targets = targets.split(',')
        variant = _parse_exclusion(variant)
        targets = [_parse_exclusion(target) for target in targets]
        groups[variant] = targets
    return groups


class BruteForceVariantPeptideCaller():
    """ Variant peptide caller using the brute force algorithm. """
    def __init__(self, reference_data:params.ReferenceData=None,
            cleavage_params:params.CleavageParams=None,
            variant_pool:seqvar.VariantRecordPool=None,
            canonical_peptides=None, tx_id:str=None,
            tx_model:gtf.TranscriptAnnotationModel=None,
            tx_seq:dna.DNASeqRecordWithCoordinates=None, gene_seq:dna.DNASeqRecord=None,
            start_index:int=None, variant_peptides:Set[str]=None,
            w2f:bool=False, selenocysteine_termination:bool=False):
        """ Constructor """
        self.reference_data = reference_data
        self.cleavage_params = cleavage_params
        self.variant_pool = variant_pool or VariantRecordPool()
        self.canonical_peptides = canonical_peptides
        self.tx_id = tx_id
        self.tx_model = tx_model
        self.tx_seq = tx_seq
        self.gene_seq = gene_seq
        self.start_index = start_index
        self.variant_peptides = variant_peptides or set()
        self.selenocysteine_termination = selenocysteine_termination
        self.w2f = w2f

    def create_canonical_peptide_pool(self):
        """ Create canonical peptide pool. """
        proteome = self.reference_data.proteome
        par = self.cleavage_params
        self.canonical_peptides = proteome.create_unique_peptide_pool(
            anno=self.reference_data.anno,
            rule=par.enzyme,
            exception=par.exception,
            miscleavage=par.miscleavage,
            min_mw=par.min_mw,
            min_length=par.min_length,
            max_length=par.max_length
        )

    def get_gene_seq(self) -> dna.DNASeqRecord:
        """ Get the gene sequence and cache it if it is not already cached. """
        if self.gene_seq:
            return self.gene_seq
        gene_id = self.tx_model.gene_id
        gene_model = self.reference_data.anno.genes[gene_id]
        chrom = gene_model.chrom
        self.gene_seq = gene_model.get_gene_sequence(self.reference_data.genome[chrom])
        return self.gene_seq

    def load_relevant_variants(self, pool:VariantRecordPool):
        """ Load relevant variants. Relevant variants are those associated with
        the transcript of `self.tx_id`, or the donor transcript of any fusion. """
        if self.tx_id not in pool:
            return

        for variant in pool[self.tx_id].transcriptional:
            self.variant_pool.add_transcriptional_variant(variant)
        for variant in pool[self.tx_id].intronic:
            self.variant_pool.add_intronic_variant(variant)
        for variant in pool[self.tx_id].fusion:
            self.variant_pool.add_fusion_variant(variant)
            right_tx_id = variant.attrs['ACCEPTER_TRANSCRIPT_ID']
            if right_tx_id not in pool:
                continue
            for v in pool[right_tx_id].transcriptional:
                self.variant_pool.add_transcriptional_variant(v)
            for v in pool[right_tx_id].intronic:
                self.variant_pool.add_intronic_variant(v)
        for variant in pool[self.tx_id].circ_rna:
            self.variant_pool.add_circ_rna(variant)

    def get_variants(self, tx_id:str, start:int, end:int,
            variant_ids:Iterable[str]=None) -> List[seqvar.VariantRecord]:
        """ Load variant records associated with the particular transcript. """
        series = self.variant_pool[tx_id]
        variants = []
        for variant in series.transcriptional:
            if variant.location.start < start -1:
                continue
            if self.tx_model.is_mrna_end_nf() and variant.location.end <= end:
                continue
            if variant_ids:
                if variant.id not in variant_ids:
                    continue
            if variant.location.start == start - 1:
                variant.to_end_inclusion(self.tx_seq)
            variants.append(variant)
        return variants

    def get_start_index(self):
        """ Get the "start index" used for filtering variants. For noncoding
        transcript, the "start index" is set to the very beginning. """
        if self.tx_seq.orf:
            self.start_index = self.tx_seq.orf.start + 3
        else:
            self.start_index = 3

    def peptide_is_valid(self, peptide:str, denylist:List[str], check_canonical) -> bool:
        """ Check whether the peptide is valid """
        if check_canonical \
                and self.canonical_peptides \
                and peptide in self.canonical_peptides:
            return False
        if peptide in denylist:
            return False
        min_len = self.cleavage_params.min_length
        max_len = self.cleavage_params.max_length
        min_mw = self.cleavage_params.min_mw
        return min_len <= len(peptide) <= max_len \
            and SeqUtils.molecular_weight(peptide, 'protein') >= min_mw

    def is_stop_lost(self, variant:seqvar.VariantRecord, reading_frame_index:int) -> bool:
        """ Check whether the variant is a stop lost mutation. """
        orf_index = self.start_index % 3

        if self.tx_model.is_protein_coding:
            if orf_index != reading_frame_index:
                return False
            orf_end = self.tx_seq.orf.end
            stop_codon = FeatureLocation(orf_end, orf_end + 3)
            return variant.location.overlaps(stop_codon)

        i = variant.variant.location.start - (variant.location.start - orf_index) % 3
        while i < variant.location.end:
            if self.tx_seq.seq[i:i+3] in ['TAA', 'TAG', 'TGA']:
                return True
            i += 3
        return False

    def should_clip_trailing_nodes(self, variants:List[seqvar.VariantRecordWithCoordinate]):
        """ Checks whether the trailing nodes should be excluded """
        return any(v.variant.is_circ_rna() for v in variants) \
            or self.tx_model.is_mrna_end_nf()

    def get_effective_variants(self, lhs:int, rhs:int,
            lrange:Tuple[int,int], rrange:Tuple[int,int], cds_start:int,
            variants:List[seqvar.VariantRecordWithCoordinate],
            variants_stop_lost:List[Tuple[bool,bool,bool]],
            variants_stop_gain:List[Tuple[bool,bool,bool]],
            variants_silent_mutation:List[Tuple[bool,bool,bool]]
            ) -> List[VariantRecordWithCoordinate]:
        """ Check whether the given range of the transcript has any variant
        associated. """
        effective_variants:List[seqvar.VariantRecordWithCoordinate] = []
        offset = 0
        query = FeatureLocation(start=lhs, end=rhs)
        start_loc = FeatureLocation(start=cds_start, end=cds_start + 3)
        rf_index = cds_start % 3
        frames_shifted = 0
        upstream_indels:List[VariantRecordWithCoordinate] = []
        is_coding = self.tx_model.is_protein_coding \
            and not any(v.variant.is_circ_rna() for v in variants)
        for i, variant_coordinate in enumerate(variants):
            variant = variant_coordinate.variant
            loc = variant_coordinate.location
            if variant.type == 'Insertion':
                loc = FeatureLocation(start=loc.start+1, end=loc.end)
            if loc.start > rhs + 3:
                break
            is_start_gain = start_loc.overlaps(loc)
            is_frameshifting = cds_start < loc.start < lhs and variant.is_frameshifting()
            if cds_start < loc.start < lhs:
                frames_shifted = (frames_shifted + variant.frames_shifted()) % 3
                if variant.is_frameshifting() \
                        and not (variant.is_fusion() or variant.is_circ_rna()):
                    upstream_indels.append(variant_coordinate)
            is_cleavage_gain = (
                loc.overlaps(FeatureLocation(start=lrange[0], end=lrange[1]))
                or loc.overlaps(FeatureLocation(start=rrange[0], end=rrange[1]))
            ) \
                if cds_start != lhs \
                else loc.overlaps(FeatureLocation(start=rhs, end=rhs + 3))

            is_stop_lost = variants_stop_lost[i][cds_start % 3] \
                and cds_start < variants[i].location.start < rhs

            is_silent_mutation = variants_silent_mutation[i][cds_start % 3] \
                and variants[i].location.start > cds_start

            is_stop_gain = variants_stop_gain[i][cds_start % 3] \
                and int((variants[i].location.start - rf_index) / 3) \
                    == int((rhs - rf_index)/3)

            if (loc.overlaps(query) \
                        or is_start_gain \
                        or (is_frameshifting and not is_coding) \
                        or is_cleavage_gain \
                        or is_stop_lost \
                        or is_stop_gain ) \
                    and not is_silent_mutation:
                effective_variants.append(variant_coordinate)
            offset += len(variant.alt) - len(variant.ref)

        if frames_shifted > 0:
            for v in upstream_indels:
                if v.variant.is_frameshifting():
                    effective_variants.append(v)
        else:
            # If there are multiple frameshifts but the overall frames shifted is 0,
            # then checks if the reference sequence skipped has any stop codon.
            if len(upstream_indels) > 1:
                first_start = upstream_indels[0].variant.location.start
                last_end = upstream_indels[-1].variant.location.end
                rf_index = cds_start % 3
                if self.has_any_stop_codon_between(first_start, last_end, rf_index):
                    effective_variants.extend(upstream_indels)
        effective_variants.sort(key=lambda v: v.location)
        return effective_variants

    def has_any_stop_codon_between(self, start:int, end:int, rf_index:int) -> bool:
        """ Checks if there is any stop codon between a given range of the tx. """
        aa_start = math.floor((start - rf_index) / 3)
        aa_end = math.ceil((end - rf_index) / 3)
        aa_end = min(aa_end, math.floor((len(self.tx_seq) - rf_index)/3))
        aa_seq:Seq = self.tx_seq.seq[rf_index:].translate()[aa_start:aa_end]
        i = 0
        sec_sites = {int(x.start) for x in self.tx_seq.selenocysteine}
        while i > -1:
            i = aa_seq.find('*', i)
            if i == -1:
                return False
            if aa_start * 3 + rf_index + i * 3 not in sec_sites:
                return True
            i += 1
        return False

    @staticmethod
    def has_overlapping_variants(variants:List[VariantRecord]) -> bool:
        """ Checks if any variants overlap. """
        for i, left in enumerate(variants):
            if i == len(variants) - 1:
                continue
            for right in variants[i+1:]:
                if left.location.end >= right.location.start:
                    return True
        return False

    def has_any_invalid_variants_on_inserted_sequences(self,
            pool:seqvar.VariantRecordPool) -> bool:
        """ Checks if any variants carried by fusion or alt splicing. In valid
        variants are those not in the region of sequence introduced by fusion.
        For example, a variant of the donor transcript before the breakpoint is
        invalid.
        """
        alt_splices = [x for x in pool[self.tx_id].transcriptional if x.is_alternative_splicing()]
        if not pool[self.tx_id].fusion and not alt_splices:
            return False

        inserted_intronic_region:Dict[str, List[FeatureLocation]] = {}

        if pool[self.tx_id].fusion:
            fusion = pool[self.tx_id].fusion[0]
            left_insert_start = fusion.attrs['LEFT_INSERTION_START']
            left_insert_end = fusion.attrs['LEFT_INSERTION_END']
            right_insert_start = fusion.attrs['RIGHT_INSERTION_START']
            right_insert_end = fusion.attrs['RIGHT_INSERTION_END']
            right_tx_id = fusion.attrs['ACCEPTER_TRANSCRIPT_ID']
            right_tx_model = self.reference_data.anno.transcripts[right_tx_id]
            right_gene_id = right_tx_model.gene_id
            right_breakpoint_gene = fusion.get_accepter_position()
            right_breakpoint_tx = self.reference_data.anno.coordinate_gene_to_transcript(
                index=right_breakpoint_gene,
                gene=right_gene_id,
                transcript=right_tx_id
            )

            if right_tx_id in pool:
                alt_splices += [x for x in pool[right_tx_id].transcriptional
                    if x.is_alternative_splicing()]

            if any(x.location.end >= fusion.location.start
                    for x in pool[self.tx_id].transcriptional):
                return True

            if left_insert_start:
                loc = FeatureLocation(start=left_insert_start, end=left_insert_end)
                if self.tx_id not in inserted_intronic_region:
                    inserted_intronic_region[self.tx_id] = []
                inserted_intronic_region[self.tx_id].append(loc)

            if right_insert_start:
                loc = FeatureLocation(start=right_insert_start, end=right_insert_end)
                if right_tx_id not in inserted_intronic_region:
                    inserted_intronic_region[right_tx_id] = []
                inserted_intronic_region[right_tx_id].append(loc)

            if right_tx_id in pool \
                    and any(x.location.start <= right_breakpoint_tx
                        for x in pool[right_tx_id].transcriptional):
                return True

        for alt_splice in alt_splices:
            donor_start = alt_splice.attrs.get('DONOR_START')
            if not donor_start:
                continue
            donor_start = int(donor_start)
            donor_end = int(alt_splice.attrs['DONOR_END'])
            loc = FeatureLocation(start=donor_start, end=donor_end)
            if alt_splice.transcript_id not in inserted_intronic_region:
                inserted_intronic_region[alt_splice.transcript_id] = []
            inserted_intronic_region[alt_splice.transcript_id].append(loc)

        if not inserted_intronic_region:
            return False

        for tx_id in pool:
            if not pool[tx_id].intronic:
                continue
            if tx_id not in inserted_intronic_region:
                return True
            for v in pool[tx_id].intronic:
                if not any(x.start < v.location.start and x.end > v.location.end
                        for x in inserted_intronic_region[tx_id]):
                    return True
        return False

    @staticmethod
    def has_any_invalid_variant_on_circ() -> bool:
        """ Checks if any variants are invlid with circRNA. """
        return False

    def has_incompatible_variants(self, pool:seqvar.VariantRecordPool) -> bool:
        """ Whether there are incompatible variants, i.e. variants that overlap
        with each other. """
        # check if there is any variants
        if self.tx_id not in pool:
            return True

        if not pool[self.tx_id].transcriptional\
                and not pool[self.tx_id].circ_rna \
                and not pool[self.tx_id].fusion:
            return True

        # check if there are multiple novel transcript variants (fusion + circ)
        n_fusion = len(pool[self.tx_id].fusion)
        n_circ = len(pool[self.tx_id].circ_rna)
        n_alt_splice = len([x for x in pool[self.tx_id].transcriptional
            if x.type in constant.ALTERNATIVE_SPLICING_TYPES])

        orf = self.tx_seq.orf
        is_mrna_end_nf = self.tx_model.is_mrna_end_nf()
        start_index = orf.start + 3 if bool(orf) else 3

        if n_circ == 0:
            for variant in pool[self.tx_id].transcriptional:
                if variant.location.start < start_index:
                    return True
                if is_mrna_end_nf and orf is not None:
                    orf_end_trinuc = FeatureLocation(start=orf.end-3, end=orf.end)
                    if variant.location.overlaps(orf_end_trinuc):
                        return True

        for fusion in pool[self.tx_id].fusion:
            if fusion.location.start < start_index:
                return True
            accepter_tx_id = fusion.attrs['ACCEPTER_TRANSCRIPT_ID']
            if accepter_tx_id not in pool:
                continue
            n_alt_splice += len([x for x in pool[accepter_tx_id].transcriptional
                if x.type in constant.ALTERNATIVE_SPLICING_TYPES])
        if n_fusion + n_circ > 1:
            return True

        # not allowing any alternative splicing comb with fusion or circ
        if (n_fusion > 0 or n_circ > 0) and n_alt_splice > 0:
            return True

        if self.has_any_invalid_variants_on_inserted_sequences(pool):
            return True

        # if self.has_any_invalid_variant_on_circ(pool):
        #     return True

        for series in pool.data.values():
            if self.has_overlapping_variants(series.transcriptional):
                return True
            if self.has_overlapping_variants(series.intronic):
                return True

        if self.tx_model.is_mrna_end_nf():
            for variant in pool[self.tx_id].transcriptional:
                if variant.location.end >= self.tx_seq.orf.end:
                    return True
                if n_circ == 0 \
                        and variant.location.start >= self.tx_seq.orf.end - 3:
                    return True
        return False

    @staticmethod
    def find_prev_cds_start_same_frame(cds_start:int, cds_start_positions:List[int]):
        """ find he previous cds start site in the same reading frame. """
        if cds_start == 0:
            return -1
        reading_frame_index = cds_start % 3
        for site in cds_start_positions[::-1]:
            if site >= cds_start:
                continue
            if site % 3 == reading_frame_index:
                return site
        return -1

    def get_variant_sequence(self, seq:Seq, location:FeatureLocation,
            offset:int, variants:List[seqvar.VariantRecord],
            pool:seqvar.VariantRecordPool
            ) -> Tuple[Seq, List[seqvar.VariantRecordWithCoordinate]]:
        """ Get variant sequence. """
        var_seq = seq
        variant_coordinates = []
        local_offset = 0
        for variant in variants:
            if variant.is_alternative_splicing():
                if variant.type == 'Deletion':
                    start = variant.location.start + local_offset - location.start
                    end = variant.location.end + local_offset - location.start
                    loc = FeatureLocation(start=start, end=start + 1)
                    variant_coordinate = seqvar.VariantRecordWithCoordinate(
                        variant=variant,
                        location=loc
                    )

                    variant_coordinates.append(variant_coordinate)
                    alt_seq = var_seq[start:start+1]
                    local_offset = local_offset + len(alt_seq) - len(variant.location)
                    var_seq = var_seq[:start] + alt_seq + var_seq[end:]

                elif variant.type == 'Insertion':
                    start = variant.location.start + local_offset - location.start
                    end = variant.location.end + local_offset - location.start

                    gene_seq = self.get_gene_seq()
                    donor_start = variant.get_donor_start()
                    donor_end = variant.get_donor_end()
                    alt_seq = str(gene_seq.seq[donor_start:donor_end])
                    loc = FeatureLocation(start=donor_start, end=donor_end)
                    insert_variants = [x for x in pool[self.tx_id].intronic
                        if loc.is_superset(x.location)]
                    alt_seq, insert_variants = self.get_variant_sequence(
                        seq=alt_seq, location=loc, offset=start,
                        variants=insert_variants, pool=pool
                    )

                    variant_coordinate = seqvar.VariantRecordWithCoordinate(
                        variant=variant,
                        location=FeatureLocation(start=start, end=start + len(alt_seq) + 1)
                    )
                    variant_coordinates.append(variant_coordinate)
                    variant_coordinates += insert_variants
                    local_offset = local_offset + len(alt_seq) + 1 - len(variant.location)
                    var_seq = var_seq[:start+1] + alt_seq + var_seq[end:]

                elif variant.type == 'Substitution':
                    start = variant.location.start + local_offset - location.start
                    end = variant.location.end + local_offset - location.start

                    gene_seq = self.get_gene_seq()
                    donor_start = variant.get_donor_start()
                    donor_end = variant.get_donor_end()
                    alt_seq = str(gene_seq.seq[donor_start:donor_end])
                    loc = FeatureLocation(start=donor_start, end=donor_end)
                    insert_variants = [x for x in pool[self.tx_id].intronic
                        if loc.is_superset(x.location)]
                    alt_seq, insert_variants = self.get_variant_sequence(
                        seq=alt_seq, location=loc, offset=start,
                        variants=insert_variants, pool=pool
                    )

                    variant_coordinate = seqvar.VariantRecordWithCoordinate(
                        variant=variant,
                        location=FeatureLocation(start=start, end=start + len(alt_seq))
                    )
                    variant_coordinates.append(variant_coordinate)
                    variant_coordinates += insert_variants
                    local_offset = local_offset + len(alt_seq) - len(variant.location)
                    var_seq = var_seq[:start] + alt_seq + var_seq[end:]

            else:
                start = variant.location.start + local_offset - location.start
                end = variant.location.end + local_offset - location.start
                loc = FeatureLocation(
                    start=start + offset, end=start + len(variant.alt) + offset
                )
                variant_coordinate = seqvar.VariantRecordWithCoordinate(
                    variant=variant,
                    location=loc
                )

                variant_coordinates.append(variant_coordinate)
                local_offset = local_offset + len(variant.alt) - len(variant.ref)
                var_seq = var_seq[:start] + variant.alt + var_seq[end:]

        return var_seq, variant_coordinates

    def get_variant_sequence_circ_rna(self, seq:Seq, variants:seqvar.VariantRecordPool
            ) -> Tuple[Seq, List[seqvar.VariantRecordWithCoordinate]]:
        """ Get the variant sequence of circRNA """
        number_of_circ = len(variants[self.tx_id].circ_rna)
        if number_of_circ != 1:
            raise ValueError(
                f"Should have exactly 1 circRNA, but {number_of_circ} were found."
            )
        circ = variants[self.tx_id].circ_rna[0]
        var_seq = Seq('')
        vars_coord = []

        for fragment in circ.fragments:
            loc = FeatureLocation(
                start=int(fragment.location.start), end=int(fragment.location.end)
            )
            frag_loc = FeatureLocation(
                start=int(fragment.location.start) + 3, end=int(fragment.location.end)
            )
            fragment = SeqFeature(location=frag_loc, chrom=fragment.chrom,
                attributes=fragment.attributes)
            new_seq = seq[loc.start:loc.end]
            frag_vars = variants.filter_variants(
                tx_ids=[circ.transcript_id], exclude_type=constant.ALTERNATIVE_SPLICING_TYPES,
                intron=False, segments=[fragment]
            )
            frag_seq, frag_vars_coord = self.get_variant_sequence(
                seq=new_seq, location=loc, offset=len(var_seq),
                variants=frag_vars, pool=variants
            )
            var_seq += frag_seq
            vars_coord += frag_vars_coord

        location = FeatureLocation(
            seqname=circ.gene_id,
            start=min(x.location.start for x in circ.fragments),
            end=max(x.location.end for x in circ.fragments)
        )
        circ_var = seqvar.VariantRecord(
            location=location,
            ref=var_seq[0],
            alt='<circRNA>',
            _type='circRNA',
            _id=circ.id
        )

        vars_aloop = copy.deepcopy(vars_coord)
        seq_aloop = copy.deepcopy(var_seq)

        for _ in range(3):
            offset = len(var_seq)
            vars_extend = copy.deepcopy(vars_aloop)
            for variant in vars_extend:
                location = FeatureLocation(
                    seqname=variant.location.seqname,
                    start=variant.location.start + offset,
                    end=variant.location.end + offset
                )
                variant.location = location
            var_seq += seq_aloop
            vars_coord += vars_extend

        circ_var_coord = VariantRecordWithCoordinate(
            variant=circ_var, location=FeatureLocation(start=0, end=len(var_seq))
        )
        vars_coord.insert(0, circ_var_coord)

        return var_seq, vars_coord

    def get_variant_sequence_fusion(self, seq:Seq, variants:seqvar.VariantRecordPool
            ) -> Tuple[Seq, List[seqvar.VariantRecordWithCoordinate]]:
        """ Get variant sequence with fusion. """
        number_of_fusion = len(variants[self.tx_id].fusion)
        if not number_of_fusion == 1:
            raise ValueError(
                f"Should have exactly 1 fusion, but {number_of_fusion} were found."
            )
        fusion = variants[self.tx_id].fusion[0]
        var_seq = seq[:fusion.location.start]
        location = FeatureLocation(start=0, end=len(var_seq))
        var_seq, variant_coordinates = self.get_variant_sequence(
            seq=var_seq, location=location, offset=0,
            variants=variants[self.tx_id].transcriptional, pool=variants
        )

        left_insert_start = fusion.attrs['LEFT_INSERTION_START']
        left_insert_end = fusion.attrs['LEFT_INSERTION_END']
        right_insert_start = fusion.attrs['RIGHT_INSERTION_START']
        right_insert_end = fusion.attrs['RIGHT_INSERTION_END']
        right_tx_id = fusion.attrs['ACCEPTER_TRANSCRIPT_ID']
        right_gene_id = fusion.attrs['ACCEPTER_GENE_ID']

        additional_seq = Seq('')
        additional_variants:List[VariantRecordWithCoordinate] = []

        # left insertion
        if left_insert_start is not None:
            gene_seq = self.get_gene_seq()
            location = FeatureLocation(start=left_insert_start, end=left_insert_end)
            insert_seq = gene_seq.seq[left_insert_start:left_insert_end]
            insert_variants = [x for x in variants[self.tx_id].intronic
                if location.is_superset(x.location)]
            insert_seq, insert_variants = self.get_variant_sequence(
                seq=insert_seq, location=location,
                offset=len(var_seq) + len(additional_seq),
                variants=insert_variants, pool=variants
            )
            location=FeatureLocation(
                start=len(var_seq),
                end=len(var_seq) + len(insert_seq)
            )
            insert_fusion = VariantRecordWithCoordinate(
                variant=fusion, location=location
            )
            insert_variants.insert(0, insert_fusion)
            additional_seq += insert_seq
            additional_variants += insert_variants

        # right insertion
        if right_insert_start is not None:
            gene_model = self.reference_data.anno.genes[right_gene_id]
            chrom = gene_model.chrom
            gene_seq = gene_model.get_gene_sequence(self.reference_data.genome[chrom])
            location = FeatureLocation(start=right_insert_start, end=right_insert_end)
            insert_seq = gene_seq.seq[right_insert_start:right_insert_end]
            insert_seq, insert_variants = self.get_variant_sequence(
                seq=insert_seq, location=location,
                offset=len(var_seq) + len(additional_seq),
                variants=variants[right_tx_id].intronic if right_tx_id in variants else [],
                pool=variants
            )
            location=FeatureLocation(
                start=len(var_seq) + len(additional_seq),
                end=len(var_seq) + len(additional_seq) + len(insert_seq)
            )
            insert_fusion = VariantRecordWithCoordinate(
                variant=fusion, location=location
            )
            insert_variants.insert(0, insert_fusion)
            additional_seq += insert_seq
            additional_variants += insert_variants

        right_tx_model = self.reference_data.anno.transcripts[right_tx_id]
        accepter_chrom = right_tx_model.transcript.location.seqname
        breakpoint_gene = fusion.get_accepter_position()
        breakpoint_tx = self.reference_data.anno.coordinate_gene_to_transcript(
            index=breakpoint_gene,
            gene=right_gene_id,
            transcript=right_tx_id
        )
        accepter_seq = right_tx_model.get_transcript_sequence(
            self.reference_data.genome[accepter_chrom]
        )
        location = FeatureLocation(start=breakpoint_tx, end=len(accepter_seq))
        insert_seq = accepter_seq.seq[breakpoint_tx:]
        insert_seq, insert_variants = self.get_variant_sequence(
            seq=insert_seq, location=location,
            offset=len(var_seq) + len(additional_seq),
            variants=variants[right_tx_id].transcriptional if right_tx_id in variants else [],
            pool=variants
        )

        location = FeatureLocation(
            start=len(var_seq) + len(additional_seq),
            end=len(var_seq) + len(additional_seq) + len(insert_seq)
        )
        fusion_var = VariantRecordWithCoordinate(variant=fusion, location=location)
        insert_variants.insert(0, fusion_var)

        additional_seq += insert_seq
        additional_variants += insert_variants

        var_seq += additional_seq
        variant_coordinates += additional_variants

        return var_seq, variant_coordinates

    def get_variant_ref_seq(self, variant:seqvar.VariantRecord) -> Seq:
        """ Get the reference sequence of a variant """
        if variant.type in ['Deletion', 'Substitution']:
            return self.tx_seq.seq[variant.location.start:variant.location.end]
        return variant.ref

    def check_variant_effect(self, seq:str,
            variants:List[seqvar.VariantRecordWithCoordinate]
            ) -> List[Tuple[bool, bool, bool]]:
        """ Check the variant effects, including stop lost, stop gain, and
        silent mutation """
        stop_lost:List[Tuple[bool, bool, bool]] = []
        stop_gain:List[Tuple[bool, bool, bool]] = []
        silent_mutation:List[Tuple[bool, bool, bool]] = []
        for variant in variants:
            skip_stop_lost = False
            skip_stop_gain = False
            skip_silent_mutation =False
            stop_gain_i:List[bool] = []
            stop_lost_i:List[bool] = []
            silent_mutation_i:List[bool] = []

            if variant.variant.is_fusion() or variant.variant.is_circ_rna():
                stop_gain_i = (False, False, False)
                skip_stop_gain = True

            if variant.variant.is_fusion() \
                    or variant.variant.is_circ_rna() \
                    or variant.variant.transcript_id != self.tx_id \
                    or 'TRANSCRIPT_ID' in variant.variant.attrs:
                stop_lost_i = (False, False, False)
                skip_stop_lost = True

            if variant.variant.is_fusion() \
                    or variant.variant.is_circ_rna() \
                    or variant.variant.is_alternative_splicing()\
                    or variant.variant.is_frameshifting():
                silent_mutation_i = (False, False, False)
                skip_silent_mutation = True

            for i in range(3):
                lhs_offset = (variant.location.start - i) % 3
                alt_lhs = variant.location.start - lhs_offset
                var_seq = seq[alt_lhs:variant.location.end]
                n_carry_over = (3 - (len(var_seq) % 3)) % 3
                alt_rhs = min(len(seq), variant.location.end + n_carry_over)
                var_seq += seq[variant.location.end:alt_rhs]

                loc = variant.variant.location
                lhs = loc.start - lhs_offset
                rhs = loc.end + (3 - (loc.end - lhs) % 3) % 3
                rhs = min(len(self.tx_seq), rhs)
                ref_seq = self.tx_seq.seq[lhs:rhs]
                ref_rf_index = lhs % 3

                var_aa = Seq(var_seq).translate(to_stop=False)
                ref_aa = Seq(ref_seq).translate(to_stop=False)
                for sec in self.tx_seq.selenocysteine:
                    if lhs <= sec.start < sec.end <= rhs \
                            and (sec.start - lhs) % 3 == 0:
                        sec_aa = int((sec.start - lhs) / 3)
                        ref_aa = ref_aa[:sec_aa] + 'U' + ref_aa[sec_aa+1:]

                    # this is when the variant is a deletion, the position is
                    # after the 3rd nt of selenocysteine codon, the codon sequence
                    # is unchanged and the amino acid sequence should be replaced
                    # to U from *.
                    if variant.variant.is_deletion():
                        if not variant.variant.is_end_inclusion() \
                                and lhs_offset == 2 \
                                and alt_lhs == sec.start and sec.end == alt_rhs \
                                and (sec.start - alt_lhs) % 3 == 0:
                            sec_aa = int((sec.start - alt_lhs) / 3)
                            var_aa = var_aa[:sec_aa] + 'U' + var_aa[sec_aa+1:]

                if not skip_stop_lost:
                    is_stop_lost = '*' not in var_aa and '*' in ref_aa \
                        and (self.tx_seq.orf is None
                            or ref_rf_index == self.tx_seq.orf.start % 3)
                    stop_lost_i.append(is_stop_lost)

                sec_altering = any(loc.overlaps(sec) for sec in self.tx_seq.selenocysteine)

                if not skip_stop_gain:
                    is_stop_gain = var_aa == '' or var_aa.startswith('*') \
                        and (
                            not (ref_aa == '' or ref_aa.startswith('*'))
                            or sec_altering
                        )
                    stop_gain_i.append(is_stop_gain)

                if not skip_silent_mutation:
                    is_silent_mutation = ref_aa == var_aa and not sec_altering
                    silent_mutation_i.append(is_silent_mutation)
            stop_lost.append(tuple(stop_lost_i))
            stop_gain.append(tuple(stop_gain_i))
            silent_mutation.append(tuple(silent_mutation_i))
        return stop_lost, stop_gain, silent_mutation

    def get_sec_positions(self, variants:List[VariantRecordWithCoordinate]) -> List[int]:
        """ Get Sec positions in the altered sequence. """
        sec_iter = iter(self.tx_seq.selenocysteine)
        var_iter = iter(variants)
        sec_i = next(sec_iter, None)
        var_i = next(var_iter, None)

        sec_positions = []
        offset = 0
        fusion_breakpoint = None
        while sec_i:
            if var_i:
                # Not consider Sec after fusion breakpoint.
                if var_i.variant.is_fusion():
                    fusion_breakpoint = var_i.variant.location.start
                    var_i = None
                    continue
                if var_i.variant.location.end <= sec_i.start:
                    ref_len = var_i.variant.get_ref_len()
                    alt_len = var_i.variant.get_alt_len()
                    offset += (alt_len - ref_len)
                    var_i = next(var_iter, None)
                    continue

                if var_i.variant.type == 'Deletion':
                    # Because for Deletion, the sequence after the REF
                    # nucleotide is deleted so the first nucleotide is unchanged.
                    var_loc = FeatureLocation(
                        var_i.variant.location.start+1,
                        var_i.variant.location.end
                    )
                else:
                    var_loc = var_i.variant.location

                if var_loc.overlaps(sec_i):
                    sec_i = next(sec_iter, None)
                    continue
            if fusion_breakpoint and sec_i.end >= fusion_breakpoint:
                break
            sec_start = sec_i.start + offset
            sec_positions.append(sec_start)
            sec_i = next(sec_iter, None)
        return sec_positions

    def call_peptides_main(self, variants:seqvar.VariantRecordPool,
            denylist:Set[str], check_variants:bool, check_canonical:bool,
            selenocysteine_termination:bool=False, is_mrna_end_nf:bool=False):
        """ Call peptide main """
        variant_peptides = set()
        tx_model = self.tx_model
        tx_seq = self.tx_seq
        rule = self.cleavage_params.enzyme
        exception = self.cleavage_params.exception

        is_circ_rna = False
        is_fusion = False

        is_coding = tx_model.is_protein_coding

        if variants[self.tx_id].fusion:
            is_fusion = True
            seq, variant_coordinates = self.get_variant_sequence_fusion(
                seq=tx_seq.seq, variants=variants
            )
        elif variants[self.tx_id].circ_rna:
            is_circ_rna = True
            gene_seq = self.get_gene_seq()
            seq, variant_coordinates = self.get_variant_sequence_circ_rna(
                seq=gene_seq.seq, variants=variants
            )
        else:
            location = FeatureLocation(start=0, end=len(tx_seq.seq))
            seq, variant_coordinates = self.get_variant_sequence(
                seq=tx_seq.seq, location=location, offset=0,
                variants=variants[self.tx_id].transcriptional, pool=variants
            )

        sec_positions = [] if is_circ_rna else \
            self.get_sec_positions(variant_coordinates)
        variant_effects = self.check_variant_effect(str(seq), variant_coordinates)
        stop_lost, stop_gain, silent_mutation = variant_effects

        if not (is_coding and is_mrna_end_nf):
            cur_cds_end = len(seq)

        if not is_coding or is_circ_rna:
            alt_seq = dna.DNASeqRecord(seq)
            cds_start_positions = alt_seq.find_all_start_codons()
            if is_fusion:
                fusion_var = None
                # Get the last fusion, which should be the actual fusion
                # no insertions of intronic regions.
                for variant in reversed(variant_coordinates):
                    if variant.variant.is_fusion():
                        fusion_var = variant
                        break
                cds_start_positions = [x for x in cds_start_positions
                    if x <= fusion_var.location.start]
        else:
            cds_start = tx_seq.orf.start
            cds_start_positions:List[int] = [cds_start]

        cds_start_sites = set(cds_start_positions)


        for cds_start in cds_start_positions:
            cur_cds_end = len(seq) - (len(seq) - cds_start) % 3

            aa_seq = seq[cds_start:cur_cds_end].translate(to_stop=False)
            if not is_circ_rna:
                for sec_start in sec_positions:
                    if (sec_start - cds_start) % 3 == 0:
                        sec_start_aa = int((sec_start - cds_start) / 3)
                        aa_seq = aa_seq[:sec_start_aa] + 'U' + aa_seq[sec_start_aa+1:]
            stop_index = aa_seq.find('*')
            if stop_index > -1:
                aa_seq = aa_seq[:stop_index]
            aa_seq = aa.AminoAcidSeqRecord(seq=aa_seq)

            sites = aa_seq.find_all_enzymatic_cleave_sites_with_ranges(rule, exception)

            # Finding the next M, so peptides that starts from the next M should
            # be processed with the correct `cds_start`
            if (not is_coding or is_circ_rna):
                next_m = aa_seq.seq[1:].find('M') + 1
                if cds_start + next_m * 3 not in cds_start_sites:
                    next_m = 0
            else:
                next_m = 0

            sites.insert(0, (0, (0,1)))
            sites.append((len(aa_seq), (len(aa_seq), len(aa_seq))))
            for j, site in enumerate(sites[:-1]):
                lhs, lrange = site
                if 0 < next_m < lhs:
                    break
                last_m = aa_seq.seq[:lhs].rfind('M')
                if last_m > 0 and not is_coding:
                    actual_cds_start = cds_start + last_m * 3
                else:
                    actual_cds_start = cds_start

                # prev_cds_start = self.find_prev_cds_start_same_frame(
                #     cds_start=actual_cds_start,
                #     cds_start_positions=cds_start_positions
                # )

                for k in range(j + 1, min([j + 3, len(sites) - 1]) + 1):
                    rhs, rrange = sites[k]
                    tx_lhs = cds_start + lhs * 3
                    tx_rhs = cds_start + rhs * 3
                    if is_mrna_end_nf and tx_rhs + 3 > len(seq):
                        continue
                    peptide = aa_seq.seq[lhs:rhs]
                    is_in_denylist = str(peptide) in denylist \
                        and (not lhs == 0 or str(peptide[1:]) in denylist)
                    if is_in_denylist:
                        continue
                    tx_lrange = (cds_start + lrange[0] * 3, cds_start + lrange[1] * 3)
                    if k == len(sites) - 1 \
                            and tx_rhs + 3 <= len(seq) \
                            and seq[tx_rhs:tx_rhs + 3].translate() == '*':
                        tx_rrange = (tx_rhs, tx_rhs + 3)
                    else:
                        tx_rrange = (cds_start + rrange[0] * 3, cds_start + rrange[1] * 3)
                    if check_variants:
                        effective_variants = self.get_effective_variants(
                            lhs=tx_lhs, rhs=tx_rhs, lrange=tx_lrange, rrange=tx_rrange,
                            cds_start=actual_cds_start, variants=variant_coordinates,
                            variants_stop_lost=stop_lost, variants_stop_gain=stop_gain,
                            variants_silent_mutation=silent_mutation
                        )

                        if not effective_variants:
                            continue
                        if is_fusion and \
                                not any(v.variant.is_fusion() for v in effective_variants):
                            continue
                    else:
                        effective_variants = []

                    peptide_seqs = self.translational_modification(
                        peptide, lhs, tx_lhs, effective_variants, denylist,
                        check_variants, check_canonical, selenocysteine_termination
                    )
                    for peptide_seq in peptide_seqs:
                        variant_peptides.add(peptide_seq)
        return variant_peptides

    def translational_modification(self, seq:Seq, lhs:int, tx_lhs:int,
            effective_variants:List[VariantRecordWithCoordinate],
            denylist:List[str], check_variants:bool, check_canonical:bool,
            selenocysteine_termination:bool
            ) -> Iterable[str]:
        """ Apply any modification that could happen during translation. """
        candidates = []
        is_start = lhs == 0 and seq.startswith('M')
        if effective_variants or not check_variants:
            candidates.append(seq)
            if is_start:
                candidates.append(seq[1:])

        if selenocysteine_termination:
            k = 0
            while k > -1:
                k = seq.find('U', k)
                if k == -1:
                    break
                if not check_variants or \
                        any(v.location.start < tx_lhs + k * 3 for v in effective_variants):
                    candidates.append(seq[:k])
                    if is_start:
                        candidates.append(seq[1:k])
                k += 1

        # W > F
        if self.w2f:
            for candidate in copy.copy(candidates):
                w2f_indices = []
                i = 0
                while i > -1:
                    i = candidate.find('W', start=i)
                    if i > -1:
                        w2f_indices.append(i)
                        i += 1
                        if i > len(candidate):
                            break

                for k in range(1, len(w2f_indices) + 1):
                    for comb in combinations(w2f_indices, k):
                        seq_mod = candidate
                        for i in comb:
                            seq_mod_new = seq_mod[:i] + 'F'
                            if i + 1 < len(candidate):
                                seq_mod_new += seq_mod[i+1:]
                            seq_mod = seq_mod_new
                        candidates.append(seq_mod)

        for candidate in candidates:
            if self.peptide_is_valid(candidate, denylist, check_canonical):
                yield str(candidate)

    def generate_variant_comb(self, fusion:bool, circ_rna:bool
            ) -> Iterable[seqvar.VariantRecordPool]:
        """ Generate combination of variants. """
        variant_type_mapper:Dict[str, Tuple[seqvar.VariantRecord, str]] = {}
        start_index = self.tx_seq.orf.start + 3 if bool(self.tx_seq.orf) else 3
        for variant in self.variant_pool[self.tx_id].transcriptional:
            if variant.location.start == start_index - 1 \
                    and (variant.is_insertion() or variant.is_deletion()) \
                    and not variant.is_alternative_splicing():
                variant.to_end_inclusion(self.tx_seq)

            variant_type_mapper[variant.id] = (variant, 'transcriptional')
        for variant in self.variant_pool[self.tx_id].intronic:
            variant_type_mapper[variant.id] = (variant, 'intronic')
        if fusion:
            for variant in self.variant_pool[self.tx_id].fusion:
                if variant.location.start < start_index - 1:
                    continue
                variant_type_mapper[variant.id] = (variant, 'fusion')
                accepter_tx_id = variant.attrs['ACCEPTER_TRANSCRIPT_ID']
                if accepter_tx_id not in self.variant_pool:
                    continue
                accepter_var_series = self.variant_pool[accepter_tx_id]
                for accepter_var in accepter_var_series.transcriptional:
                    variant_type_mapper[accepter_var.id] = (accepter_var, 'transcriptional')
                for accepter_var in accepter_var_series.intronic:
                    variant_type_mapper[accepter_var.id] = (accepter_var, 'intronic')
        if circ_rna:
            for variant in self.variant_pool[self.tx_id].circ_rna:
                variant_type_mapper[variant.id] = (variant, 'circ_rna')

        all_variants = [v[0] for v in variant_type_mapper.values()]

        for i in range(len(all_variants)):
            for inds in combinations(range(len(all_variants)), i + 1):
                variants = [all_variants[i] for i in inds]
                if fusion \
                        and not any(variant_type_mapper[v.id][1] == 'fusion'
                                    for v in variants):
                    continue
                if circ_rna \
                        and not any(variant_type_mapper[v.id][1] == 'circ_rna'
                                    for v in variants):
                    continue
                pool = seqvar.VariantRecordPool()
                pool.anno = self.variant_pool.anno
                for variant in variants:
                    var_type = variant_type_mapper[variant.id][1]
                    tx_id = variant.transcript_id
                    if var_type == 'transcriptional':
                        pool.add_transcriptional_variant(variant, tx_id)
                    elif var_type == 'intronic':
                        pool.add_intronic_variant(variant, tx_id)
                    elif var_type == 'fusion':
                        pool.add_fusion_variant(variant, tx_id)
                    elif var_type == 'circ_rna':
                        pool.add_circ_rna(variant, tx_id)
                pool.sort()
                if self.has_incompatible_variants(pool):
                    continue
                yield pool


    def call_peptides(self):
        """ Call variant peptides """
        empty_pool = seqvar.VariantRecordPool()
        empty_pool[self.tx_id] = seqvar.TranscriptionalVariantSeries()
        denylist = self.call_peptides_main(
            variants=empty_pool, denylist=set(),
            check_variants=False, check_canonical=False,
            selenocysteine_termination=self.selenocysteine_termination,
            is_mrna_end_nf=False
        )
        main_peptides = set()

        for comb in self.generate_variant_comb(fusion=False, circ_rna=False):
            peptides = self.call_peptides_main(
                variants=comb, denylist=denylist,
                check_variants=True,  check_canonical=True,
                selenocysteine_termination=self.selenocysteine_termination,
                is_mrna_end_nf=self.tx_model.is_mrna_end_nf()
            )
            self.variant_peptides.update(peptides)
            main_peptides.update(peptides)

        for comb in self.generate_variant_comb(fusion=True, circ_rna=False):
            donor_tx_id = comb[self.tx_id].fusion[0].accepter_transcript_id
            donor_tx_model = self.reference_data.anno.transcripts[donor_tx_id]
            peptides = self.call_peptides_main(
                variants=comb, denylist=denylist,
                check_variants=True, check_canonical=True,
                is_mrna_end_nf=donor_tx_model.is_mrna_end_nf()
            )
            self.variant_peptides.update(peptides)

        denylist.update(main_peptides)
        for comb in self.generate_variant_comb(fusion=False, circ_rna=True):
            peptides = self.call_peptides_main(
                variants=comb, denylist=denylist,
                check_variants=True, check_canonical=True,
                is_mrna_end_nf=True
            )
            self.variant_peptides.update(peptides)

def create_mnvs(pool:seqvar.VariantRecordPool, max_adjacent_as_mnv:int
        ) -> seqvar.VariantRecordPool:
    """ Create MNVs """
    for tx_id in pool.data.keys():
        mnvs = seqvar.find_mnvs_from_adjacent_variants(
            pool[tx_id].transcriptional,
            max_adjacent_as_mnv
        )
        pool[tx_id].transcriptional = sorted(pool[tx_id].transcriptional + mnvs)

        mnvs = seqvar.find_mnvs_from_adjacent_variants(
            pool[tx_id].intronic,
            max_adjacent_as_mnv
        )
        pool[tx_id].intronic = sorted(pool[tx_id].intronic + mnvs)
    return pool

def fix_indel_after_start_codon(pool:seqvar.VariantRecordPool,
        ref:params.ReferenceData) -> seqvar.VariantRecordPool:
    """ Fix indel variants that are right after the start codon by shifting
    it to end inclusion to be consistant with callVariant """
    for tx_id in pool.data.keys():
        tx_model = ref.anno.transcripts[tx_id]
        chrom = tx_model.transcript.chrom
        tx_seq = tx_model.get_transcript_sequence(ref.genome[chrom])

        if tx_seq.orf:
            start = tx_seq.orf.start
        else:
            start = 0
        for v in pool[tx_id].transcriptional:
            if v.location.start == start + 2 \
                    and (v.is_insertion() or v.is_deletion()) \
                    and not v.is_fusion() \
                    and not v.is_alternative_splicing():
                v.to_end_inclusion(tx_seq)
    return pool

def main(args):
    """ main """
    # Load genomic references
    anno, genome, proteome = load_references(
        path_anno=args.reference_dir/'annotation.gtf',
        path_genome=args.reference_dir/'genome.fasta',
        path_proteome=args.reference_dir/'proteome.fasta'
    )
    reference_data = params.ReferenceData(
        genome=genome,
        anno=anno,
        proteome=proteome,
        canonical_peptides=None
    )

    # load GVF files
    variant_pool = seqvar.VariantRecordPool()
    variant_pool.anno = reference_data.anno
    for path in args.input_gvf:
        with open(path) as handle:
            variant_pool.load_variants(
                handle=handle,
                anno=reference_data.anno,
                genome=reference_data.genome
            )
    variant_pool = fix_indel_after_start_codon(variant_pool, reference_data)
    variant_pool = create_mnvs(variant_pool, args.max_adjacent_as_mnv)
    variant_peptides:Set[str] = set()
    for tx_id in variant_pool.data.keys():
        caller = BruteForceVariantPeptideCaller()
        caller.variant_pool = variant_pool
        caller.reference_data = reference_data
        caller.tx_id = tx_id
        caller.tx_model = caller.reference_data.anno.transcripts[caller.tx_id]
        caller.tx_seq = caller.tx_model.get_transcript_sequence(
            caller.reference_data.genome[caller.tx_model.transcript.chrom]
        )
        caller.w2f = args.w2f_reassignment
        caller.selenocysteine_termination = args.selenocysteine_termination

        caller.cleavage_params = params.CleavageParams(
            enzyme=args.cleavage_rule,
            exception=args.cleavage_exception,
            miscleavage=int(args.miscleavage),
            min_mw=float(args.min_mw),
            min_length=args.min_length,
            max_length=args.max_length
        )

        caller.create_canonical_peptide_pool()
        caller.get_start_index()

        caller.call_peptides()

        variant_peptides.update(caller.variant_peptides)

    for peptide in sorted(variant_peptides):
        print(peptide, file=sys.stdout)



""" GTFPointer and GTFPointerDict """
from __future__ import annotations
from typing import Dict, Set, Iterable, Mapping, Union, IO
from collections import deque
from moPepGen.gtf import GtfIO
from moPepGen.gtf.GeneAnnotationModel import GeneAnnotationModel
from moPepGen.gtf.TranscriptAnnotationModel import (
    TranscriptAnnotationModel,
    GTF_FEATURE_TYPES
)
from moPepGen.gtf.GTFSourceInferrer import GTFSourceInferrer


GENE_DICT_CACHE_SIZE = 10
TX_DICT_CACHE_SIZE = 10

class GTFPointer():
    """ GTFPointer. Represents the range of the GTF file of a particular
    genome annotation entity (gene or transcript). """
    def __init__(self, handle:IO, key:str, start:int, end:int, source:str):
        """ """
        self.handle = handle
        self.key = key
        self.start = start
        self.end = end
        self.source = source

    def __len__(self) -> int:
        """ length """
        return self.end - self.start

class GenePointer(GTFPointer):
    """ Pointer of a GTF file for a gene. """
    def __init__(self, handle: IO, key: str, start: int, end: int, source: str,
            transcripts:Set[str]=None):
        """ Constructor """
        super().__init__(handle, key, start, end, source)
        self.transcripts = transcripts or set()

    def load(self) -> GeneAnnotationModel:
        """ Load gene annotation model from handle. """
        cur = self.handle.tell()
        offset = self.start - cur
        self.handle.seek(offset, 1)
        buffer:bytes = self.handle.read(len(self))
        lines = buffer.decode('utf-8').rstrip().split('\n')
        if len(lines) > 1:
            raise ValueError(f"Multiple lines found for gene {self.key}")
        record = GtfIO.line_to_seq_feature(lines[0])
        record.__class__ = GeneAnnotationModel
        record.exons = []
        record.transcripts = list(self.transcripts)
        return record

    def to_line(self) -> str:
        """ Convert to line """
        fields = [
            self.key,
            str(self.start),
            str(self.end),
            ','.join(self.transcripts)
        ]
        return '\t'.join(fields)

class TranscriptPointer(GTFPointer):
    """ Pointer of a GTF file for a trnascript. """
    def __init__(self, handle: IO, key: str, start: int, end: int, source: str,
            is_protein_coding:bool=None):
        """ TranscriptPointer """
        super().__init__(handle, key, start, end, source)
        self.is_protein_coding = is_protein_coding

    def load(self) -> TranscriptAnnotationModel:
        """ Load pointer and return a transcript annotation model. """
        cur = self.handle.tell()
        offset = self.start - cur
        self.handle.seek(offset, 1)
        buffer:bytes = self.handle.read(len(self))
        lines = buffer.decode('utf-8').rstrip().split('\n')
        tx_model = TranscriptAnnotationModel()

        for line in lines:
            record = GtfIO.line_to_seq_feature(line)
            feature = record.type.lower()
            if feature not in GTF_FEATURE_TYPES:
                continue
            tx_id = record.transcript_id
            if tx_id != self.key:
                raise ValueError("Loaded transcript do not match.")
            record.id = tx_id
            tx_model.add_record(feature, record)
        tx_model.sort_records()
        return tx_model

    def to_line(self) -> str:
        """ Convert to string """
        fields = [
            self.key,
            str(self.start),
            str(self.end),
            str(self.is_protein_coding)
        ]
        return "\t".join(fields)

def iterate_pointer(handle:IO, source:str=None) -> Iterable[Union[GenePointer, TranscriptPointer]]:
    """ Iterate over a GTF file and yield pointers. """
    if not source:
        inferrer = GTFSourceInferrer()

    cur_gene_id:str = None
    cur_tx_id:str = None
    cur_gene_pointer = None
    cur_tx_pointer = None
    line_end = 0
    for line in handle:
        line:bytes
        line_start = line_end
        line_end += len(line)
        line = line.decode('utf-8')

        if line.startswith('#'):
            continue

        record = GtfIO.line_to_seq_feature(line)

        if not source:
            record.source = inferrer.infer(record)
        else:
            record.source = source

        if record.type.lower() == 'gene':
            if cur_gene_pointer:
                yield cur_gene_pointer
            if cur_tx_pointer:
                yield cur_tx_pointer
            cur_gene_id = record.gene_id
            cur_gene_pointer = GenePointer(
                handle, cur_gene_id,
                line_start, line_end, record.source
            )
            cur_tx_id = None
            cur_tx_pointer = None
        else:
            if cur_tx_id != record.transcript_id:
                if cur_tx_pointer:
                    yield cur_tx_pointer
                cur_tx_id = record.transcript_id
                cur_tx_pointer = TranscriptPointer(
                    handle, cur_tx_id,
                    line_start, line_end, record.source
                )
            else:
                cur_tx_pointer.end = line_end
            if cur_gene_pointer:
                cur_gene_pointer.transcripts.add(cur_tx_id)
    if cur_gene_pointer:
        yield cur_gene_pointer
    if cur_tx_pointer:
        yield cur_tx_pointer

class GTFPointerDict(dict, Mapping[str, GTFPointer]):
    """ GTFPointerDict. This defines a special dict class that keys are str
    (gene or transcript ID) and values are corresponding GTFPointer. The getter
    magic function is customized so it returns a loaded GeneAnnotationModel
    or TranscriptAnnotationModel instead of the pointer.
    """
    def __init__(self, *args:Dict[str,GTFPointer], **kwargs:Dict[str,GTFPointer]):
        """ Constructor """
        for val in kwargs.values():
            self.validate(val, GTFPointer)
        if args:
            if not isinstance(args[0], dict):
                raise TypeError('Data type invalid')
            for val in args[0].values():
                self.validate(val, GTFPointer)
        super().__init__(*args, **kwargs)

    def __iter__(self) -> Iterable[str]:
        """ Iterator """
        for k in self.keys():
            yield k

    def validate(self, val, __type):
        """ Ensures the values are GTFPointers """
        if not __type is GTFPointer or not issubclass(__type, GTFPointer):
            raise TypeError(f"type = {__type} is not supported")
        if not isinstance(val, __type):
            raise TypeError(
                f"{self.__class__} only accecpts {__type.__class__} objects"
            )

    def __setitem__(self, __key: str, __value: GTFPointer) -> None:
        """ Setter """
        self.validate(__value, GTFPointer)
        return super().__setitem__(__key, __value)

    def get_pointer(self, __key: str) -> GTFPointer:
        """ Get pointer by key without loading the annotation entity. """
        return super().__getitem__(__key)

class GenePointerDict(GTFPointerDict):
    """ GenePointerDict. This defines a dict class that keys are gene IDs and
    values are the corresponding gene models. The getter magic function is
    customized so it returns a loaded GeneAnnotationModel object instead of the
    pointer.
    """
    def __init__(self, *args: Dict[str, GenePointer], **kwargs: Dict[str, GenePointer]):
        """ Constructor """
        for val in kwargs.values():
            self.validate(val, GenePointer)
        if args:
            if not isinstance(args[0], dict):
                raise TypeError('Data type invalid')
            for val in args[0].values():
                self.validate(val, GenePointer)
        super().__init__(*args, **kwargs)
        self._cache = {}
        self._cached_keys = deque()

    def __getitem__(self, __key: str) -> GeneAnnotationModel:
        """ Getter """
        if __key in self._cache:
            return self._cache[__key]
        self._cached_keys.appendleft(__key)
        if len(self._cached_keys) > GENE_DICT_CACHE_SIZE:
            key_pop = self._cached_keys.pop()
            self._cache.pop(key_pop)
        pointer:GenePointer = self.get_pointer(__key)
        val = pointer.load()
        self._cache[__key] = val
        return val

class TranscriptPointerDict(GTFPointerDict):
    """ TranscriptPointerDict. This defines a dict class that keys are transcript
    IDs and values are the corresponding transcript models. The getter magic
    function is customized so it returns a loaded TranscriptAnnotationModel
    object instead of the pointer.
    """
    def __init__(self, *args: Dict[str, TranscriptPointer], **kwargs: Dict[str, TranscriptPointer]):
        """ Constructor """
        for val in kwargs.values():
            self.validate(val, TranscriptPointer)
        if args:
            if not isinstance(args[0], dict):
                raise TypeError('Data type invalid')
            for val in args[0].values():
                self.validate(val, TranscriptPointer)
        super().__init__(*args, **kwargs)
        self._cache = {}
        self._cached_keys = deque()

    def __getitem__(self, __key: str) -> TranscriptAnnotationModel:
        """ getter """
        if __key in self._cache:
            return self._cache[__key]
        self._cached_keys.appendleft(__key)
        if len(self._cached_keys) > TX_DICT_CACHE_SIZE:
            key_pop = self._cached_keys.pop()
            self._cache.pop(key_pop)
        pointer:TranscriptPointer = self.get_pointer(__key)
        val = pointer.load()
        val.is_protein_coding = pointer.is_protein_coding is True
        val.transcript.source = pointer.source
        self._cache[__key] = val
        return val



""" Module for Amino Acid """
from moPepGen.aa.AminoAcidSeqRecord import AminoAcidSeqRecord,\
    AminoAcidSeqRecordWithCoordinates
from moPepGen.aa.AminoAcidSeqDict import AminoAcidSeqDict
from moPepGen.aa.VariantPeptidePool import VariantPeptidePool
from moPepGen.aa.PeptidePoolSplitter import PeptidePoolSplitter, \
    VariantSourceSet, VariantPeptideInfo
from moPepGen.aa.VariantPeptideIdentifier import create_variant_peptide_id, \
    parse_variant_peptide_id



""" Frameshifting mutations of a node """
from __future__ import annotations
import copy
from typing import FrozenSet, List, Set
from moPepGen import seqvar


class VariantCombinations():
    """ Node Framesfhits """
    def __init__(self, data:Set[FrozenSet[seqvar.VariantRecord]]=None, max_n:int=3):
        """ """
        self.data = data or set()
        self.max_n = max_n

    def add(self, variant:seqvar.VariantRecord) -> None:
        """ Combine a variant to each of the variant combinations """
        data = set()
        for comb in self.data:
            if len(comb) < self.max_n:
                comb = set(comb)
                comb.add(variant)
                comb = frozenset(comb)
            data.add(comb)
        self.data = data

    def extend(self, variants:VariantCombinations) -> None:
        """ Extend another VariantCombinations  """
        new_one = copy.copy(variants)
        self.data.update(new_one.data)

    def __copy__(self) -> VariantCombinations:
        """ create a copy """
        data = set()
        for comb in self.data:
            data.add(copy.copy(comb))
        return self.__class__(data)

    def remove(self, variant:seqvar.VariantRecord):
        """ remove """
        data = set()
        for comb in self.data:
            if variant in comb:
                comb = set(comb)
                comb.remove(variant)
                if len(comb) > 0:
                    comb = frozenset(comb)
                    data.add(comb)
            else:
                data.add(comb)
        self.data = data

    def update(self, variants:List[seqvar.VariantRecordWithCoordinate],
            add_singleton:bool=True) -> None:
        """ For a given VariantRecordWithCoordinates, update those that are
        frameshift mutations """
        for variant in variants:
            if variant.variant.is_frameshifting():
                self.add(variant.variant)
            if add_singleton:
                singleton = frozenset([variant.variant])
                if singleton not in self.data:
                    self.data.add(singleton)



""" `indexGVF`  """
from __future__ import annotations
import argparse
from pathlib import Path
import tempfile
from moPepGen import get_logger, check_sha512
from moPepGen.seqvar import GVFIndex
from moPepGen.seqvar import GVFMetadata
from moPepGen.cli import common


INPUT_FILE_FORMATS = ['.gvf']

# pylint: disable=W0212
def add_subparser_index_gvf(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen indexGVF """
    p:argparse.ArgumentParser = subparsers.add_parser(
        name='indexGVF',
        help='Generate index for GVF file.',
        description='Generate index for GVF file.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=p, formats=INPUT_FILE_FORMATS,
        message='File path to GVF file. Must be generated by any of the'
        ' moPepGen parsers。'
    )
    common.add_args_debug_level(p)
    p.set_defaults(func=index_gvf)
    common.print_help_if_missing_args(p)
    return p


def index_gvf(args:argparse.Namespace):
    """ Generate GVF index """
    input_file:Path = args.input_path
    common.validate_file_format(input_file, INPUT_FILE_FORMATS, True)
    output_file = input_file.with_suffix(input_file.suffix + '.idx')
    with open(input_file, 'rb') as handle:
        sum_val = check_sha512(handle)

    common.print_start_message(args)

    with open(input_file, 'rt') as handle:
        metadata = GVFMetadata.parse(handle)

    is_circ_rna = metadata.is_circ_rna()

    with open(input_file, 'rb') as in_handle:
        with tempfile.TemporaryFile(mode='w+t') as temp_file:
            temp_file.write(f"# CHECKSUM={sum_val}\n")
            in_handle.seek(0)
            it = GVFIndex.iterate_pointer(handle=in_handle, is_circ_rna=is_circ_rna)
            for pointer in it:
                temp_file.write(pointer.to_line() + '\n')
            temp_file.seek(0)
            with open(output_file, 'wt') as out_handle:
                for line in temp_file:
                    out_handle.write(line)

    get_logger().info('GVF index generated.')



""" Module for rMATS parser """
from __future__ import annotations
from typing import TYPE_CHECKING, List, Tuple
from moPepGen import gtf, dna, seqvar
from .RMATSRecord import RMATSRecord


if TYPE_CHECKING:
    from moPepGen.seqvar import SpliceJunction


class A3SSRecord(RMATSRecord):
    """ Alternative 3' Splicing Site """
    def __init__(self, gene_id:str, gene_symbol:str, chrom:str,
            long_exon_start:int, long_exon_end:int, short_exon_start:int,
            short_exon_end:int, flanking_exon_start, flanking_exon_end:int,
            ijc_sample_1:int, sjc_sample_1:int, ijc_sample_2:int,
            sjc_sample_2:int, inc_form_len:int, skip_form_len:int,
            pvalue:float, fdr:float):
        """ Constructor """
        super().__init__(gene_id, gene_symbol, chrom)
        self.long_exon_start = long_exon_start
        self.long_exon_end = long_exon_end
        self.short_exon_start = short_exon_start
        self.short_exon_end = short_exon_end
        self.flanking_exon_start = flanking_exon_start
        self.flanking_exon_end = flanking_exon_end
        self.ijc_sample_1 = ijc_sample_1
        self.sjc_sample_1 = sjc_sample_1
        self.ijc_sample_2 = ijc_sample_2
        self.sjc_sample_2 = sjc_sample_2
        self.inc_form_len = inc_form_len
        self.skip_form_len = skip_form_len
        self.pvalue = pvalue
        self.fdr = fdr

    @classmethod
    def readline(cls, line:str) -> A3SSRecord:
        """ Create a A5SSRecord from a line of the rMASTS output """
        fields = line.rstrip().split('\t')
        return cls(
            gene_id=fields[1].strip('"'),
            gene_symbol=fields[2].strip('"'),
            chrom=fields[3],
            long_exon_start=int(fields[5]),
            long_exon_end=int(fields[6]),
            short_exon_start=int(fields[7]),
            short_exon_end=int(fields[8]),
            flanking_exon_start=int(fields[9]),
            flanking_exon_end=int(fields[10]),
            ijc_sample_1=int(fields[12]),
            sjc_sample_1=int(fields[13]),
            ijc_sample_2=None if fields[14] == '' else int(fields[14]),
            sjc_sample_2=None if fields[15] == '' else int(fields[15]),
            inc_form_len=int(fields[16]),
            skip_form_len=int(fields[17]),
            pvalue=None if fields[18] == 'NA' else float(fields[18]),
            fdr=None if fields[19] == 'NA' else float(fields[19])
        )

    def create_variant_id(self, anno:gtf.GenomicAnnotation) -> str:
        """ Create variant ID """
        if anno.genes[self.gene_id].strand == 1:
            uee = anno.coordinate_genomic_to_gene(self.flanking_exon_end - 1, self.gene_id) + 1
            les = anno.coordinate_genomic_to_gene(self.long_exon_start, self.gene_id)
            ses = anno.coordinate_genomic_to_gene(self.short_exon_start, self.gene_id)
        else:
            uee = anno.coordinate_genomic_to_gene(self.flanking_exon_start, self.gene_id) + 1
            les = anno.coordinate_genomic_to_gene(self.long_exon_end - 1, self.gene_id)
            ses = anno.coordinate_genomic_to_gene(self.short_exon_end - 1, self.gene_id)

        return f"A5SS_{uee}-{les}-{ses}"

    def create_splice_junctions(self, strand:int) -> Tuple[SpliceJunction, SpliceJunction]:
        """ Create splice junctions """
        if strand == 1:
            long_junction = seqvar.SpliceJunction(
                upstream_start=self.flanking_exon_start,
                upstream_end=self.flanking_exon_end,
                downstream_start=self.long_exon_start,
                downstream_end=self.long_exon_end,
                gene_id=self.gene_id,
                chrom=self.chrom
            )
            short_junction = seqvar.SpliceJunction(
                upstream_start=self.flanking_exon_start,
                upstream_end=self.flanking_exon_end,
                downstream_start=self.short_exon_start,
                downstream_end=self.short_exon_end,
                gene_id=self.gene_id,
                chrom=self.chrom
            )
        else:
            long_junction = seqvar.SpliceJunction(
                upstream_start=self.long_exon_start,
                upstream_end=self.long_exon_end,
                downstream_start=self.flanking_exon_start,
                downstream_end=self.flanking_exon_end,
                gene_id=self.gene_id,
                chrom=self.chrom
            )
            short_junction = seqvar.SpliceJunction(
                upstream_start=self.short_exon_start,
                upstream_end=self.short_exon_end,
                downstream_start=self.flanking_exon_start,
                downstream_end=self.flanking_exon_end,
                gene_id=self.gene_id,
                chrom=self.chrom
            )
        return long_junction, short_junction

    def convert_to_variant_records(self, anno:gtf.GenomicAnnotation,
            genome:dna.DNASeqDict, min_ijc:int, min_sjc:int
            ) -> List[seqvar.VariantRecord]:
        """ Convert A3SS record to VariantRecord """
        variants = []

        gene_model = anno.genes[self.gene_id]
        strand = gene_model.strand

        long_junction, short_junction = self.create_splice_junctions(strand)

        if not long_junction.is_novel(anno) and not short_junction.is_novel(anno):
            return variants

        tx_ids = gene_model.transcripts
        chrom = gene_model.location.seqname
        gene_seq = gene_model.get_gene_sequence(genome[chrom])
        var_id = self.create_variant_id(anno)

        if gene_model.strand == 1:
            upstream_novel = False
            downstream_novel = True
        else:
            upstream_novel = True
            downstream_novel = False

        # For A3SS, long is inclusion
        for tx_id in tx_ids:
            tx_model = anno.transcripts[tx_id]

            if self.ijc_sample_1 >= min_ijc:
                aln = long_junction.align_to_transcript(
                    tx_model, upstream_novel, downstream_novel
                )
                if aln:
                    variants += aln.convert_to_variant_records(anno, gene_seq, var_id)

            if self.sjc_sample_1 >= min_sjc:
                aln = short_junction.align_to_transcript(
                    tx_model, upstream_novel, downstream_novel
                )
                if aln:
                    variants += aln.convert_to_variant_records(anno, gene_seq, var_id)
        return variants



""" Common functions for cli """
from __future__ import annotations
import argparse
import os
import sys
from typing import Tuple, Set, List, TYPE_CHECKING
from pathlib import Path
import errno
import signal
import functools
import time
import logging
import pkg_resources
from moPepGen import aa, dna, gtf, seqvar, get_logger, constant
from moPepGen.aa.expasy_rules import EXPASY_RULES
from moPepGen.index import IndexDir


if TYPE_CHECKING:
    from moPepGen.params import CleavageParams

def print_help_if_missing_args(parser:argparse.ArgumentParser):
    """ If no args are provided, print help and exit """
    if len(sys.argv) == 2 and sys.argv[1] == parser.prog.split(' ')[-1]:
        parser.print_help(sys.stderr)
        sys.exit(1)

def print_start_message(args:argparse.Namespace):
    """ Print the program start message """
    get_logger().info('moPepGen %s started', args.command)

def add_args_reference(parser:argparse.ArgumentParser, genome:bool=True,
        proteome:bool=True, index:bool=True):
    """ Add args for reference files including reference genome assembly FASTA,
    annotation GTF, proteome FASTA, and indexed files. """
    group = parser.add_argument_group('Reference Files')
    if genome:
        group.add_argument(
            '-g', '--genome-fasta',
            type=Path,
            help='Path to the genome assembly FASTA file. Only ENSEMBL and'
            ' GENCODE are supported. Its version must be the same as the'
            ' annotation GTF and proteome FASTA',
            metavar='<file>',
            default=None
        )
    group.add_argument(
        '-a', '--annotation-gtf',
        type=Path,
        help='Path to the annotation GTF file. Only ENSEMBL and GENCODE are'
        ' supported. Its version must be the same as the genome and proteome'
        ' FASTA.',
        metavar='<file>',
        default=None
    )
    group.add_argument(
        '--reference-source',
        type=str,
        choices=['GENCODE', 'ENSEMBL'],
        help='Source of reference genome and annotation.',
        default=None
    )
    if proteome:
        group.add_argument(
            '-p', '--proteome-fasta',
            type=Path,
            help='Path to the translated protein sequence FASTA file. Only'
            ' ENSEMBL and GENCODE are supported. Its version must be the same'
            ' as genome FASTA and annotation GTF.',
            metavar='<file>',
            default=None
        )
        group.add_argument(
            '--invalid-protein-as-noncoding',
            action='store_true',
            help='Treat any transcript that the protein sequence is invalid ('
            ' contains the * symbol) as noncoding.'
        )
    if index:
        group.add_argument(
            '--index-dir',
            type=Path,
            help='Path to the directory of index files generated by moPepGen'
            ' generateIndex. If given, --genome-fasta, --proteome-fasta and'
            ' --anntotation-gtf will be ignored.',
            metavar='<file>',
            nargs='?',
            default=None
        )

def add_args_cleavage(parser:argparse.ArgumentParser, enzyme_only:bool=False):
    """ Add args for cleavage """
    group = parser.add_argument_group('Cleavage Parameters')
    group.add_argument(
        '-c', '--cleavage-rule',
        type=str,
        help='Enzymatic cleavage rule.',
        default='trypsin',
        metavar='<value>',
        choices=list(EXPASY_RULES.keys())
    )
    group.add_argument(
        '--cleavage-exception',
        type=str,
        help='Enzymatic cleavage exception.',
        default='auto',
        metavar='<value>'
    )
    if enzyme_only:
        return
    group.add_argument(
        '-m', '--miscleavage',
        type=int,
        help='Number of cleavages to allow per non-canonical peptide.',
        default=2,
        metavar='<number>'
    )
    group.add_argument(
        '-w', '--min-mw',
        type=float,
        help='The minimal molecular weight of the non-canonical peptides.',
        default=500.,
        metavar='<number>'
    )
    group.add_argument(
        '-l', '--min-length',
        type=int,
        help='The minimal length of non-canonical peptides, inclusive.',
        default=7,
        metavar='<number>'
    )
    group.add_argument(
        '-x', '--max-length',
        type=int,
        help='The maximum length of non-canonical peptides, inclusive.',
        default=25,
        metavar='<number>'
    )

def add_args_decoy(parser:argparse.ArgumentParser):
    """ add decoy fasta related arguments """
    group = parser.add_argument_group('Decoy Database Parameters')
    group.add_argument(
        '--decoy-string',
        type=str,
        default='DECOY_',
        help='The decoy string that is combined with the FASTA header for decoy'
        ' sequences.',
        metavar='<value>'
    )
    group.add_argument(
        '--decoy-string-position',
        type=str,
        choices=['prefix', 'suffix'],
        help='Should the decoy string be placed at the start or end of FASTA'
        ' headers?',
        default='prefix',
        metavar='<value>'
    )

def add_args_debug_level(parser:argparse.ArgumentParser):
    """ add debug level """
    parser.add_argument(
        '--debug-level',
        type=str,
        help='Debug level.',
        default='INFO',
        metavar='<value|number>'
    )
    parser.add_argument(
        '-q', '--quiet',
        action='store_true',
        help='Quiet',
        default=False
    )

def add_args_output_path(parser:argparse.ArgumentParser, formats:List[str]):
    """ add output file path """
    parser.add_argument(
        '-o', '--output-path',
        type=Path,
        help=f'File path to the output file. Valid formats: {formats}',
        metavar='<file>',
        required=True
    )

def add_args_input_path(parser:argparse.ArgumentParser, formats:List[str],
        plural:bool=False, message:str=None, required:bool=True):
    """ Add input file path """
    if message is None:
        message = f"File path to the input file{'s' if plural else ''}."

    message += f"{' Can take multiple files.' if plural else ''} Valid formats: " +\
        f"{formats}"

    metavar = ["<files>"] if plural else "<file>"

    if plural:
        nargs = '+' if required else '*'
    else:
        nargs = None

    parser.add_argument(
        '-i', '--input-path',
        type=Path,
        help=message,
        nargs=nargs,
        metavar=metavar,
        required=required
    )


def add_args_source(parser:argparse.ArgumentParser):
    """ Add source """
    parser.add_argument(
        '--source',
        type=str,
        help='Variant source (e.g. gSNP, sSNV, Fusion)',
        required=True
    )

def load_references(args:argparse.Namespace, load_genome:bool=True,
        load_canonical_peptides:bool=True, load_proteome:bool=False,
        invalid_protein_as_noncoding:bool=False, check_protein_coding:bool=False,
        cleavage_params:CleavageParams=None
        ) -> Tuple[dna.DNASeqDict, gtf.GenomicAnnotationOnDisk, Set[str]]:
    """ Load reference files. If index_dir is specified, data will be loaded
    from pickles, otherwise, will read from FASTA and GTF. """
    logger = get_logger()
    genome = None
    anno = None
    canonical_peptides = None
    if invalid_protein_as_noncoding:
        load_proteome = True

    if args.index_dir:
        index_dir = IndexDir(args.index_dir)
        index_dir.validate_metadata()

        if load_canonical_peptides:
            canonical_peptides = index_dir.load_canonical_peptides(cleavage_params)

        if load_genome:
            genome = index_dir.load_genome()

        anno = index_dir.load_annotation()

        if load_proteome:
            proteome = index_dir.load_proteome()

        if invalid_protein_as_noncoding:
            anno.check_protein_coding(proteome, True)

        logger.info('Reference indices loaded.')
    else:
        if (check_protein_coding is True or load_canonical_peptides is True) and \
                args.proteome_fasta is None:
            raise ValueError('--proteome-fasta was not specified.')
        anno = gtf.GenomicAnnotationOnDisk()
        anno.generate_index(args.annotation_gtf, source=args.reference_source)

        logger.info('Annotation GTF loaded.')

        if load_proteome or load_canonical_peptides or check_protein_coding:
            proteome = aa.AminoAcidSeqDict()
            proteome.dump_fasta(args.proteome_fasta, source=args.reference_source)
            logger.info('Proteome FASTA loaded.')
            anno.check_protein_coding(proteome, invalid_protein_as_noncoding)

        if load_genome:
            genome = dna.DNASeqDict()
            genome.dump_fasta(args.genome_fasta)
            logger.info('Genome assembly FASTA loaded.')

        if load_canonical_peptides:
            rule:str = args.cleavage_rule
            miscleavage:int = int(args.miscleavage)
            min_mw:float = float(args.min_mw)
            exception = args.cleavage_exception
            min_length:int = args.min_length
            max_length:int = args.max_length
            canonical_peptides = proteome.create_unique_peptide_pool(
                anno=anno, rule=rule, exception=exception,
                miscleavage=miscleavage, min_mw=min_mw, min_length=min_length,
                max_length=max_length
            )
            logger.info('canonical peptide pool generated.')

    if not load_proteome:
        proteome = None

    return genome, anno, proteome, canonical_peptides

def generate_metadata(args:argparse.Namespace) -> seqvar.GVFMetadata:
    """ Generate metadata """
    if args.index_dir:
        reference_index = args.index_dir.resolve()
        genome_fasta = None
        annotation_gtf = None
    else:
        reference_index = None
        genome_fasta = args.genome_fasta.resolve() if \
            hasattr(args, 'genome_fasta') else None
        annotation_gtf = args.annotation_gtf.resolve() if \
            hasattr(args, 'annotation_gtf') else None

    return seqvar.GVFMetadata(
        parser=args.command,
        source=args.source,
        chrom='Gene ID',
        reference_index=reference_index,
        genome_fasta=genome_fasta,
        annotation_gtf=annotation_gtf
    )

def load_inclusion_exclusion_biotypes(args:argparse.Namespace
        ) -> Tuple[List[str], List[str]]:
    """ Load inclusion and exclusion biotypes """
    inclusion_biotypes = []
    if args.inclusion_biotypes:
        with open(args.inclusion_biotypes, 'rt') as handle:
            for line in handle:
                inclusion_biotypes.append(line.rstrip())

    exclusion_path = args.exclusion_biotypes
    if not exclusion_path:
        exclusion_path = pkg_resources.resource_filename(
            'moPepGen', 'data/gencode_hs_exclusion_list.txt'
        )

    exclusion_biotypes = []
    if exclusion_path:
        with open(exclusion_path, 'rt') as handle:
            for line in handle:
                exclusion_biotypes.append(line.rstrip())

    return inclusion_biotypes, exclusion_biotypes

def parse_range(x:str) -> Tuple[int,int]:
    """ parse range from argument """
    y = x.split(',')
    if len(y) != 2:
        raise ValueError('Range invalid')
    return tuple(int(i) for i in y)

def validate_file_format(file:Path, types:List[str]=None, check_readable:bool=False,
            check_writable:bool=False, is_directory:bool=False):
    """ Validate the file type """
    if types is None:
        types = []
    if not is_directory:
        suffixes = file.suffixes
        actual_suffixes = [suffixes[-1]]
        if len(suffixes) > 1:
            actual_suffixes.append(''.join(suffixes[-2:]))
        if not any(suffix in types for suffix in actual_suffixes):
            raise ValueError(
                f'The file {file} is invalid. Valid file types are {types}.'
            )
    if check_readable:
        if not file.exists():
            raise FileNotFoundError(f"File does not exist: '{file}'")
        if not os.access(file, os.R_OK):
            raise PermissionError(f"Permission denied: '{file}'")
    if check_writable:
        if file.exists():
            if not os.access(file, os.W_OK):
                raise PermissionError(f"Permission denied: '{file}'")
        else:
            if not os.access(file.parent, os.W_OK):
                raise PermissionError(f"Permission denied: '{file}'")

# pylint: disable=unused-argument
def timeout(seconds=10, error_message=os.strerror(errno.ETIME)):
    """ Decorator to raise a TimeoutError if the process runs over time. """
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if 'timeout' in kwargs and kwargs['timeout']:
                seconds = kwargs['timeout']
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wrapper

    return decorator

def setup_loggers(level:str):
    """ Initialize loggers for both init and run """
    # pylint: disable=protected-access
    try:
        level = int(level)
    except ValueError:
        pass
    try:
        log_level = logging._checkLevel(level)
    except ValueError:
        log_level = logging._checkLevel('INFO')

    logger = logging.getLogger(constant.PROG_NAME)
    logger.setLevel(log_level)

    formatter = logging.Formatter(
        '[ %(asctime)s ] [ %(name)s - %(levelname)5s ] %(message)s',
        "%Y-%m-%d %H:%M:%S"
    )
    logging.Formatter.converter = time.localtime

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(log_level)

    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger



""" Module for cursor for codon finding """
from __future__ import annotations
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from moPepGen.svgraph.TVGNode import TVGNode

class TVGCursor():
    """ Cursor class, used when walking in the DNA transcirpt graph for
    finding codons. """
    def __init__(self, node:TVGNode, search_orf:bool=False):
        """ Constructor """
        self.node = node
        self.search_orf = search_orf



"""Main entry point"""
import sys
import argparse
from moPepGen import cli, __version__, constant, get_logger
from moPepGen.util.ResourcesMonitor import ResourcesMonitor
from moPepGen.cli.common import setup_loggers

CLI_MAIN_USAGE = f"{constant.PROG_NAME} [-h] [-V] <command> [options]"
CLI_MAIN_DESCRIPTION = """
-- Indexing
   generateIndex       Generate genome and proteome index files for moPepGen.
   updateIndex         Update moPepGen index.
   indexGVF            Generate index for GVF files.

-- Parsing
   parseVEP            Parse VEP output.
   parseREDItools      Parse REDItools annotated output.
   parseSTARFusion     Parse STAR-Fusion output.
   parseFusionCatcher  Parse FusionCatcher output.
   parseArriba         Parse Arriba output.
   parseRMATS          Parse rMATS output.
   parseCIRCexplorer   Parse CIRCexplorer known circRNA output.

-- Calling
   callVariant         Call non-canonical peptides from genomic variants.
   callNovelORF        Call non-canonical peptides from novel ORFs.
   callAltTranslation  Call non-canonital peptides with alternative translation
                       from coding transcripts.

-- Processing
   summarizeFasta      Summarize variant peptides.
   filterFasta         Filter noncanonical peptides.
   splitFasta          Split noncanonical peptides into separate databases.
   mergeFasta          Merge multiple variant peptide FASTA databases.
   encodeFasta         Encode variant peptide FASTA file header.
   decoyFasta          Generate decoy database FASTA file.
"""

def main():
    """ Main entry point """
    process_monitor = ResourcesMonitor()

    parser = argparse.ArgumentParser(
        prog=constant.PROG_NAME,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        usage=CLI_MAIN_USAGE
    )
    parser.add_argument(
        '-V', '--version',
        help='Show version number and exit.',
        action='store_true'
    )
    subparsers = parser.add_subparsers(
        dest='command',
        help=argparse.SUPPRESS,
        description=CLI_MAIN_DESCRIPTION
    )

    cli.add_subparser_generate_index(subparsers)
    cli.add_subparser_update_index(subparsers)
    cli.add_subparser_parse_vep(subparsers)
    cli.add_subparser_parse_reditools(subparsers)
    cli.add_subparser_parse_star_fusion(subparsers)
    cli.add_subparser_parse_fusion_catcher(subparsers)
    cli.add_subparser_parse_arriba(subparsers)
    cli.add_subparser_parse_rmats(subparsers)
    cli.add_subparser_parse_circexplorer(subparsers)
    cli.add_subparser_call_variant(subparsers)
    cli.add_subparser_call_novel_orf(subparsers)
    cli.add_subparser_call_alt_translation(subparsers)
    cli.add_subparser_split_fasta(subparsers)
    cli.add_subparser_filter_fasta(subparsers)
    cli.add_subparser_index_gvf(subparsers)
    cli.add_subparser_merge_fasta(subparsers)
    cli.add_subparser_encode_fasta(subparsers)
    cli.add_subparser_decoy_fasta(subparsers)
    cli.add_subparser_summarize_fasta(subparsers)

    # allowing values to start with -, such as -100,3
    # https://stackoverflow.com/a/21446783/11081630
    for i, arg in enumerate(sys.argv):
        if arg[0] == '-' and arg[1].isdigit():
            sys.argv[i] = ' ' + arg

    args = parser.parse_args()

    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    debug_level = 0 if args.quiet else args.debug_level
    setup_loggers(debug_level)
    logger = get_logger()

    if args.version:
        print(f'moPepGen {__version__}', file=sys.stdout, flush=True)
        sys.exit()

    args.func(args)

    resources_usage = process_monitor.get_resource_usage()
    logger.info(resources_usage)

if __name__ == '__main__':
    main()



""" This module defines the class for a single GTF record
"""
from __future__ import annotations
from typing import Dict


class GTFRecord():
    """ A GTFRecord object holds an entry from the a GTF file.

    The schema of GTFRecord follows the definition from ENSMEBL (https://uswest
    .ensembl.org/info/website/upload/gff.html).

    Attributes:
        seqname (str): The reference sequence name of the chromosome or
            scaffold.
        source (str): Name of the program that generated this feature, or the
            data source (database or project name).
        feature (str): Feature type name, e.g. gene, transcript, exone.
        start (int): Start position* of the feature, with sequence numbering
            starting at 1.
        end (int): End position* of the feature, with sequence numbering
            starting at 1.
        score (float): A floating point value.
        strand (str): Defined as + (forward) or - (reverse).
        frame (int): One of '0', '1' or '2'. '0' indicates that the first base
            of the feature is the first base of a codon, '1' that the second
            base is the first base of a codon, and so on.
        attributes (dict[str, str]): additional information about each feature,
            e.g., gene_id, transcript_id, gene_type.
    """
    def __init__(
            self, seqname: str, source: str, feature: str, start: int,
            end: int, score: float, strand: str, frame: int,
            attributes: Dict[str, str]):
        self.seqname = seqname
        self.source = source
        self.feature = feature
        self.start = start
        self.end = end
        self.score = score
        self.strand = strand
        self.frame = frame
        self.attributes = attributes

    @property
    def biotype(self):
        """ biotype """
        return self.attributes['gene_type']

    def __eq__(self, other:GTFRecord)->bool:
        """ equal to """
        return self.start == other.start and self.end == other.end

    def __ne__(self, other:GTFRecord)->bool:
        """ not equal """
        return not self == other

    def __gt__(self, other:GTFRecord)->bool:
        """ greater than """
        if self.start > other.start:
            return True
        if self.start < other.start:
            return False
        if self.end > other.end:
            return True
        if self.end < other.end:
            return False
        return False

    def __ge__(self, other:GTFRecord)->bool:
        """ greater or equal """
        return self > other or self == other

    def __lt__(self, other:GTFRecord)->bool:
        """ less than """
        return not (self > other or self == other)

    def __le__(self, other:GTFRecord)->bool:
        """" less or equal """
        return self < other or self == other



""" module for peptide pool splitter """
from __future__ import annotations
from typing import Dict, IO, List, Set, TYPE_CHECKING
from pathlib import Path
from moPepGen.seqvar import GVFMetadata
from moPepGen import seqvar, circ, constant, VARIANT_PEPTIDE_SOURCE_DELIMITER, \
    SPLIT_DATABASE_KEY_SEPARATER
from .VariantPeptidePool import VariantPeptidePool
from .VariantPeptideLabel import VariantPeptideInfo, VariantSourceSet, \
    LabelSourceMapping

if TYPE_CHECKING:
    from .AminoAcidSeqRecord import AminoAcidSeqRecord

Databases = Dict[str,VariantPeptidePool]

class PeptidePoolSplitter():
    """ PeptidePoolSplitter to split peptide pool into separate databases.

    Attributes:
        peptides (VariantPeptidePool): The variant peptide pool.
        order (Dict[str,str]): Order of sources.
        label_map (Dict[str,Dict[str,str]]): This a nested dict of transcripts,
            variant lables, and their corresponding source type. This is to
            quickly retrieve the source type of a given variant label.
        group_map (Dict[str,str]): A map of certain variant sources being
            grouped together.
        databases (Dict[str,VariantPeptidePool]): This is the splitted
            databases. It is a dict that keys are the string representation of
            variant sources, and values are the splitted VariantPeptidePool.
        sources (Set[str]): Variant sources.
    """
    def __init__(self, peptides:VariantPeptidePool=None, order:Dict[str,int]=None,
            label_map:LabelSourceMapping=None, group_map:Dict[str,str]=None,
            databases:Databases=None):
        self.peptides = peptides
        self.databases = databases or {}
        self.label_map = label_map or LabelSourceMapping()
        self.group_map = group_map or {}
        self.order = order or {}


    def get_reversed_group_map(self) -> Dict[str, List[str]]:
        """ Reverse group map """
        group_map: Dict[str, List[str]] = {}
        for k,v in self.group_map.items():
            if v in group_map:
                group_map[v].append(k)
            else:
                group_map[v] = [k]
        return group_map

    def append_order_internal_sources(self):
        """ Add internal sources that are not present in any GTFs, including
        novel ORF, sec termination, and codon reassignment. """
        sources = [
            constant.SOURCE_NOVEL_ORF,
            constant.SOURCE_SEC_TERMINATION,
            constant.SOURCE_CODON_REASSIGNMENT
        ]
        for source in sources:
            if source in self.group_map:
                source = self.group_map[source]
            if source not in self.order:
                self.append_order(source)

    def append_order(self, source:str):
        """ Add a group to the end of the order. """
        if source in self.order:
            return
        if source in self.group_map:
            source = self.group_map[source]
        if source in self.order:
            return
        self.order[source] = max(self.order.values()) + 1 if self.order else 0

    def load_database(self, handle:IO) -> None:
        """ load peptide database """
        pool = VariantPeptidePool.load(handle)
        if not self.peptides:
            self.peptides = pool
            return
        for peptide in pool.peptides:
            self.peptides.add_peptide(peptide, None, skip_checking=True)

    def load_gvf(self, handle:IO) -> None:
        """ Load variant lables from GVF file. """
        metadata = GVFMetadata.parse(handle)
        source = metadata.source

        if source not in self.order:
            self.append_order(source)

        if metadata.parser == 'parseCIRCexplorer':
            for record in circ.io.parse(handle):
                self.label_map.add_circ_rna(record, source)
        else:
            for variant in seqvar.io.parse(handle):
                self.label_map.add_variant(variant, source)

    def add_peptide_to_database(self, database_key:str,
            peptide:AminoAcidSeqRecord) -> None:
        """ Add peptide to database """
        if database_key not in self.databases:
            self.databases[database_key] = VariantPeptidePool()
        self.databases[database_key].peptides.add(peptide)

    def get_all_peptide_sources(self, tx2gene:Dict[str,str], coding_tx:Set[str]):
        """ Get sources of all peptides """
        peptide_sources = []
        for peptide in self.peptides.peptides:
            peptide_infos = VariantPeptideInfo.from_variant_peptide(
                peptide=peptide, tx2gene=tx2gene, coding_tx=coding_tx,
                label_map=self.label_map
            )
            peptide_infos.sort()
            peptide_sources.append([x.sources for x in peptide_infos])
        return peptide_sources

    def split(self, max_groups:int, additional_split:List[Set],
            tx2gene:Dict[str,str], coding_tx:Set[str]):
        """ Split peptide pool into separate databases """
        self.append_order_internal_sources()
        VariantSourceSet.set_levels(self.order)
        delimiter = VARIANT_PEPTIDE_SOURCE_DELIMITER
        additional_split = [VariantSourceSet(x) for x in additional_split]
        for peptide in self.peptides.peptides:
            peptide_infos = VariantPeptideInfo.from_variant_peptide(
                peptide=peptide,
                tx2gene=tx2gene,
                coding_tx=coding_tx,
                label_map=self.label_map,
                group_map=self.group_map
            )
            peptide_infos.sort()

            peptide.description = delimiter.join([str(x) for x in peptide_infos])
            peptide.id = peptide.description
            peptide.name = peptide.description

            sources = peptide_infos[0].sources

            if len(sources) <= max_groups:
                database_key = str(sources)
                self.add_peptide_to_database(database_key, peptide)
            else:
                has_additional_splitting = False
                for additional_set in additional_split:
                    if additional_set.issubset(sources):
                        database_key = self.get_additional_database_key(additional_set)
                        self.add_peptide_to_database(database_key, peptide)
                        has_additional_splitting = True
                        break
                if not has_additional_splitting:
                    database_key = self.get_remaining_database_key()
                    self.add_peptide_to_database(database_key, peptide)

    @staticmethod
    def get_additional_database_key(additional:str) -> str:
        """ Get the database key for priority """
        return f'{str(additional)}{SPLIT_DATABASE_KEY_SEPARATER}additional'

    @staticmethod
    def get_remaining_database_key() -> str:
        """ Get the database key for remanent peptides """
        return 'Remaining'

    @staticmethod
    def get_database_filename(database_key:str, prefix:str) -> str:
        """ Get the database filename """
        return f"{prefix}_{database_key}.fasta"

    def write(self, output_prefix:str) -> None:
        """ Write databases to disk """
        output_dir = Path(output_prefix).parent
        output_dir.mkdir(exist_ok=True)
        filename_prefix = str(Path(output_prefix).name)
        for database_key, database in self.databases.items():
            filename = self.get_database_filename(database_key, filename_prefix)
            database.write(output_dir/filename)


def get_index_contains_source_set(peptide_infos:List[VariantPeptideInfo],
        target_set:VariantSourceSet) -> List[int]:
    """ Find the indices of a list of source sets,  """
    for i, info in enumerate(peptide_infos):
        if target_set.issubset(info.sources):
            return i
    return None



""" Module for REDItoolsParser """
from __future__ import annotations
from typing import List, Tuple, Iterable
import re
from moPepGen.SeqFeature import FeatureLocation
from moPepGen.seqvar.VariantRecord import VariantRecord
from moPepGen import seqvar, gtf, ERROR_INDEX_IN_INTRON


def parse(path:str, transcript_id_column:int=16
        ) -> Iterable[REDItoolsRecord]:
    """ Parse the REDItools output table and returns an iterator. The input
    table must be annotated using the AnnotateTable.py script from REDItools 1.

    Args:
        path (str): Path to the REDItools output table.
        transcript_id_column (int): The column that holds the transcript_id.
            Defaults to 10.

    Return:
        A iterable of REDItoolsRecord.
    """
    with open(path, 'r') as handle:
        line = next(handle, None)
        line = next(handle, None)
        while line:
            line = line.rstrip()
            line = re.sub('[,&$]$', '', line)
            fields = line.split('\t')
            base_count = [int(i) for i in fields[6].strip('][')\
                    .split(', ')]
            all_subs = []
            for sub in fields[7].split(' '):
                if len(sub) > 2:
                    raise ValueError('length of sub larger than 2')
                all_subs.append((sub[0], sub[1]))

            try:
                g_coverage_q = int(fields[9])
            except ValueError:
                g_coverage_q = None

            if len(fields) <= transcript_id_column:
                raise ValueError('transcript_id_column invalid.')
            _data = re.split(r',|&|\$', fields[transcript_id_column])
            transcript_ids = [tuple(x.split('-')) for x in _data]

            yield REDItoolsRecord(
                region=fields[0],
                position=int(fields[1]),
                reference=fields[2],
                strand=int(fields[3]),
                coverage_q=int(fields[4]),
                mean_quality=float(fields[5]),
                base_count=base_count,
                all_subs=all_subs,
                frequency=float(fields[8]),
                g_coverage_q=g_coverage_q,
                transcript_id=transcript_ids
            )
            line = next(handle, None)


class REDItoolsRecord():
    """ A REDItoolsRecord defines the attributes from a REDItools output table.

    Attributes:
        region (str): chromosome sequence name.
        position (int): 1 based position.
        reference (str): nucleotide in reference.
        strand (int): 1 for position; -1 for negative.
        coverage_q30 (int): depth per site at quality score of 30.
        mean_quality (int): mean quality score per site.
        base_cout (List[int]): base distribution per site in the order of
            A, C, G and T
        all_subs (List[Tuple[str, str]]): all substritutions.
        frequency (int): observed frequency of substitution.
        transcirpt_id (List[Tuple[str,str]]): transcript IDs.
    """
    def __init__(self, region:str, position:int, reference:str, strand:int,
            coverage_q:int, mean_quality:float, base_count:List[int],
            all_subs:List[Tuple[str, str]], frequency:int, g_coverage_q:int,
            transcript_id:List[Tuple[str, str]]):
        """ constructor """
        self.region = region
        self.position = position
        self.reference = reference
        self.strand = strand
        self.all_subs = all_subs
        self.coverage_q = coverage_q
        self.mean_quality = mean_quality
        self.base_count = base_count
        self.all_subs = all_subs
        self.frequency = frequency
        self.g_coverage_q = g_coverage_q
        self.transcript_id = transcript_id
        self.base_count_order = {
            'A': 0,
            'C': 1,
            'G': 2,
            'T': 3
        }

    def get_valid_subs(self, min_coverage_alt:int, min_frequency_alt:float,
            min_coverage_rna:int, min_coverage_dna:int) -> bool:
        """ Get all valid substitutions. """
        total_count = sum(self.base_count)
        valid_subs = []

        if total_count < min_coverage_rna:
            return valid_subs

        if self.g_coverage_q != -1:
            if self.g_coverage_q is None or self.g_coverage_q < min_coverage_dna:
                return valid_subs

        for sub in self.all_subs:
            alt = sub[1]
            read_count = self.base_count[self.base_count_order[alt]]
            if read_count < min_coverage_alt:
                continue
            if read_count / total_count < min_frequency_alt:
                continue
            valid_subs.append(sub)
        return valid_subs

    def convert_to_variant_records(self, anno:gtf.GenomicAnnotation,
            min_coverage_alt:int, min_frequency_alt:float,
            min_coverage_rna:int, min_coverage_dna:int
            ) -> List[seqvar.VariantRecord]:
        """ Convert to VariantRecord.

        Args:
            anno (gtf.TranscriptFTFDict): the reference genome annotations.

        Return:
            A list of seqvar.VariantRecord. Multiple variant records can be
            returned, because a genomic position can be on multiple transcripts
            with different splicing.
        """
        _ids = []
        for tx_id, feature in self.transcript_id:
            if feature == 'transcript':
                _ids.append(tx_id)

        records = []
        genomic_location = f"{self.region}:{self.position}"
        for tx_id in _ids:
            tx_model:gtf.TranscriptAnnotationModel = anno.transcripts[tx_id]
            try:
                position = tx_model.get_transcript_index(self.position - 1)
            except ValueError as e:
                if e.args[0] == ERROR_INDEX_IN_INTRON:
                    continue
            gene_id = tx_model.transcript.gene_id
            gene_model = anno.genes[gene_id]
            strand = gene_model.strand
            position = anno.coordinate_genomic_to_gene(self.position - 1, gene_id)
            location = FeatureLocation(
                seqname=gene_id,
                start=position,
                end=position + 1
            )
            valid_subs = self.get_valid_subs(
                min_coverage_alt=min_coverage_alt,
                min_frequency_alt=min_frequency_alt,
                min_coverage_rna=min_coverage_rna,
                min_coverage_dna=min_coverage_dna
            )
            for sub in valid_subs:
                ref = sub[0]
                alt = sub[1]

                _id = f'RES-{position + 1}-{ref}-{alt}'
                attrs = {
                    'TRANSCRIPT_ID': tx_id,
                    'GENOMIC_POSITION': genomic_location,
                    'STRAND': strand
                }
                record = VariantRecord(
                    location=location,
                    ref=ref,
                    alt=alt,
                    _type='RNAEditingSite',
                    _id=_id,
                    attrs=attrs
                )
                records.append(record)
        return records



""" DNASeqRecord """
from __future__ import annotations
import copy
from typing import Iterable, List
import re
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from moPepGen.SeqFeature import FeatureLocation, MatchedLocation
from moPepGen.aa.expasy_rules import EXPASY_RULES
from moPepGen import aa


class DNASeqRecord(SeqRecord):
    #pylint: disable=W0223
    """ A DNASeqRecord object holds a single DNA sequence and information about
    it. Inherits the Bio.SeqRecord.SeqRecord class.

    Attributes:
        seq (Bio.Seq.Seq: The DNA sequence.
        id (str)
        name (str)
        description (str)

    For other attributes see biopython's documentation:
    https://biopython.org/docs/1.75/api/Bio.Seq.html?highlight=seq#Bio.Seq.Seq
    """

    def __add__(self, other:DNASeqRecord)->DNASeqRecord:
        """ add operation """
        new = super().__add__(other)
        new.__class__ = DNASeqRecord
        return new

    def __hash__(self):
        """ hash """
        return hash(str(self.seq))

    def reverse_complement(self, _id=False, name=False, description=False,
            features=True, annotations=False, letter_annotations=True,
            dbxrefs=False)->DNASeqRecord:
        """ Returns the reverse complement sequence """
        seq = super().reverse_complement(
            id=_id, name=name, description=description, features=features,
            annotations=annotations, letter_annotations=letter_annotations,
            dbxrefs=dbxrefs
        )
        seq.__class__ = self.__class__
        return seq

    def find_start_codon(self) -> int:
        """ find start codon index """
        return self.seq.find('ATG')

    def iter_start_codon(self) -> Iterable[int]:
        """ Create iterators of start codon positions. """
        for it in re.finditer('ATG', str(self.seq)):
            yield it.start()

    def find_stop_codon(self, start:int=0) -> int:
        """ Find and return the stop codon.

        Args:
            start (int): The start position to search. It only searches the
                reading frame specified by start.
            table (str): The codon table to choose (unsupported)

        Returns:
            The index of the stop codon. -1 is returned if not found.
        """
        # NOTE(CZ): enable choosing codon table from the Bio.Data.CodonTable
        # module.
        stop_codons = ['TAA', 'TAG', 'TGA']
        n = len(self)
        for i in range(start, n - n % 3, 3):
            if str(self.seq[i:i+3]) in stop_codons:
                return i
        return -1

    def find_orf_first(self, start:int=0) -> int:
        """ Search for the first open reading frame (start codon)

        Args:
            start (int): The starting index to search.

        Returns:
            The index of the start codon. -1 is returned if not found.
        """
        # NOTE(CZ): consider other start codons?
        seq:Seq = self.seq
        return seq.find('ATG', start=start)

    def find_orf_all(self, start:int=0) -> int:
        """ Returns iterator of start codon positions """

        while True:
            start = self.find_orf_first(start)
            if start == -1:
                return
            yield start
            start += 3

    def iter_enzymatic_cleave_sites(self, rule:str, exception:str=None,
            start:int=None, end:int=None) -> Iterable[int]:
        """ Returns a generator of the enzymatic cleave sites of a given range.
        Default is the entire sequence.

        Args:
            rule (str): The rule for enzymatic cleavage, e.g., trypsin.
            exception (str): The exception for cleavage rule.
            start (int): Index to start searching.
            end (int): Index to stop searching.

        Returns:
            A generator of cleave sites.
        """
        if start is None:
            start = 0
        if end is None:
            end = len(self)
        peptides = str(self.seq[start:end].translate())
        rule = EXPASY_RULES[rule]
        exception = EXPASY_RULES.get(exception, exception)
        exceptions = [] if exception is None else \
            [x.end() for x in re.finditer(exception, peptides)]

        for x in re.finditer(rule, peptides):
            if x not in exceptions:
                yield x.end() * 3 + start

    def find_all_enzymatic_cleave_sites(self, rule:str, exception:str=None,
            start:int=None, end:int=None) -> List[int]:
        """ Find all enzymatic cleave sites. """
        if start is None:
            start = 0
        if end is None:
            end = len(self)
        end = end - (end - start) % 3
        return list(self.iter_enzymatic_cleave_sites(rule=rule,
            exception=exception, start=start, end=end))

    def find_all_start_codons(self) -> List[int]:
        """ Find all start codon positions """
        start_positions = []
        start_codon = 'ATG'
        while True:
            if len(start_positions) == 0:
                i = 0
            else:
                i = start_positions[-1] + 1
            j = self.seq[i:].find(start_codon)
            if j == -1:
                break
            start_positions.append(i + j)
        return start_positions

    def find_last_cleave_position(self, end:int, rule:str, exception:str=None,
            miscleavage:int=0) -> int:
        """ Find the last enzymatic cleave site with a given number of
        miscleavage. """
        end = end - end % 3
        cleave_sites = [0]
        cleave_sites.extend(self.find_all_enzymatic_cleave_sites(rule=rule,
            exception=exception, start=None, end=end))
        i = len(cleave_sites) - miscleavage
        return max(i, 0)

    def find_next_cleave_position(self, start:int, rule:str,
            exception:str=None, miscleavage:int=0) -> int:
        """ Find the next enzymatic cleave site with a given number of
        miscleavage. Start must be in the correct ORF to return the correct
        site. """
        i = 0
        for site in self.iter_enzymatic_cleave_sites(rule=rule,
                exception=exception, start=start, end=None):
            if i == miscleavage:
                return site + start
            i += 1
        return -1

    def find_cleave_positions_before(self, position:int, rule:str,
            exception:str=None, miscleavage:str=2) -> List[int]:
        """ Find specified number of enzymatic cleave sites before a given
        position.
        """
        sites = []
        position = position - position % 3
        left_cleave_sites = [0]
        left_cleave_sites.extend(self.find_all_enzymatic_cleave_sites(
            rule=rule, exception=exception, start=None, end=position))
        i = miscleavage +1
        while i > 0:
            if len(left_cleave_sites) >= i:
                sites.append(left_cleave_sites[-i])
            else:
                sites.append(None)
            i += 1
        return sites

    def find_cleave_positions_after(self, position:int, rule:str,
            exception:str=None, miscleavage:str=2) -> List[int]:
        """ Find specified number of enzymatic cleave sites before a given
        position. """
        sites = []
        start = position - position % 3
        right_cleave_sites = self.find_all_enzymatic_cleave_sites(rule=rule,
            exception=exception, start=start, end=None)
        for i, _ in enumerate(sites):
            sites[i] += start
        right_cleave_sites.append(len(self))
        for i in range(miscleavage):
            if len(right_cleave_sites) >= i + 1:
                sites.append(right_cleave_sites[i])
            else:
                sites.append(None)
        return sites


    def find_cleave_positions_within(self, start:int, end:int, rule:int,
            exception:str=None) -> List[int]:
        """ Find enzymatic cleave sites within a given range """
        start = start - start % 3
        end = end + 3 - end % 3
        left = self.find_last_cleave_position(end=start, rule=rule,
            exception=exception, miscleavage=0)
        right = self.find_next_cleave_position(start=end, rule=rule,
            exception=exception, miscleavage=0)
        new_sites = self[left:right].find_all_enzymatic_cleave_sites(rule=rule,
            exception=exception)
        for i, _ in enumerate(new_sites):
            new_sites[i] += start
        return new_sites

    def translate(self, table:str='Standard', stop_symbol:str="*",
            to_stop:bool=False, cds:bool=False, gap:str="-"
            ) -> aa.AminoAcidSeqRecord:
        """ Translate the DNA sequence to a amino acid sequence.

        Args:
            table (str): Which codon table to use? This can be either a name
                (string), an NCBI identifier (integer), or a CodonTable object
                (useful for non-standard genetic codes). This defaults to the
                "Standard" table.
            to_stop (bool):  Boolean, defaults to False meaning do a full
                translation continuing on past any stop codons (translated as
                the specified stop_symbol). If True, translation is terminated
                at the first in frame stop codon (and the stop_symbol is not
                appended to the returned protein sequence).
        """
        if len(self.seq) % 3 > 0:
            end = len(self.seq) - len(self.seq) % 3
            seq:Seq = self.seq[:end]
            peptides = seq.translate(
                table=table, stop_symbol=stop_symbol, to_stop=to_stop,
                cds=cds, gap=gap
            )
        else:
            peptides = self.seq.translate(
                table=table, stop_symbol=stop_symbol, to_stop=to_stop,
                cds=cds, gap=gap
            )
        return aa.AminoAcidSeqRecord(
            seq=peptides,
            _id=self.id,
            name=self.name,
            description=self.description
        )


class DNASeqRecordWithCoordinates(DNASeqRecord):
    """ The DNASeqReocrdWithLocation holds a sequence and its location matched
    to another sequence, e.g., chromosome or transcript. It contains two
    additional attributes, locations and orf.

    Attributes:
        seq (Seq): The DNA sequence.
        locations (List[MatchedLocation]): The locations where this
            sequence is aligned tp.
        orf (FeatureLocation): The open reading frame start and end.
    """
    def __init__(self, seq:Seq, *args, locations:List[MatchedLocation]=None,
            orf:FeatureLocation=None, selenocysteine:List[FeatureLocation]=None,
            **kwargs,):
        """ Constract a DNASeqRecordWithCoordinates object.

        Args:
            seq (Seq): The DNA sequence.
            locations (List[MatchedLocation]): The locations where this
                sequence is aligned tp.
            orf (FeatureLocation): The open reading frame start and end.
        """
        super().__init__(seq=seq, *args, **kwargs)
        self.locations = locations or []
        # query index
        self.orf = orf
        self.selenocysteine = selenocysteine or []

    def __add__(self, other:DNASeqRecordWithCoordinates
            ) -> DNASeqRecordWithCoordinates:
        """ add operation """
        new = super().__add__(other)
        new.__class__ = DNASeqRecordWithCoordinates
        left_locs = copy.copy(self.locations)
        right_locs = copy.copy(other.locations)
        if left_locs and right_locs:
            lhs = self.locations[-1]
            rhs = other.locations[0].shift(len(self))
            if lhs.ref.end == rhs.ref.start and lhs.query.end == rhs.query.start \
                    and lhs.ref.seqname == rhs.ref.seqname \
                    and lhs.query.reading_frame_index == rhs.query.reading_frame_index:
                query = FeatureLocation(
                    start=lhs.query.start, end=rhs.query.end,
                    seqname=lhs.query.seqname,
                    reading_frame_index=lhs.query.reading_frame_index
                )
                ref = FeatureLocation(start=lhs.ref.start, end=rhs.ref.end,
                    seqname=lhs.ref.seqname)
                new_loc = MatchedLocation(query=query, ref=ref)
                right_locs.pop(0)
                left_locs[-1] = new_loc

        new.locations = left_locs + [loc.shift(len(self)) for loc in right_locs]
        new.orf = self.orf
        new.selenocysteine = self.selenocysteine
        return new

    def __getitem__(self, index)->DNASeqRecordWithCoordinates:
        """ get item """
        if isinstance(index, int):
            return super().__getitem__(index)
        start, stop, _ = index.indices(len(self))

        locations = []
        if start != stop:
            for location in self.locations:
                lhs = location.query.start
                rhs = location.query.end
                if rhs <= start:
                    continue
                if lhs >= stop:
                    break
                if lhs <= start:
                    if rhs <= stop:
                        location = location[start-lhs:]
                    else:
                        location = location[start-lhs:stop-lhs]
                else:
                    if rhs <= stop:
                        location = location.shift(-start)
                    else:
                        new_location = location[:stop-lhs]
                        new_location = new_location.shift(lhs-start)
                        location = new_location

                locations.append(location)

        return self.__class__(
            seq=self.seq[index],
            locations=locations,
            orf = self.orf,
            selenocysteine=self.selenocysteine,
            id=self.id,
            name=self.name,
            description=self.description
        )

    def __hash__(self):
        """hash"""
        return hash(self.seq)

    def __eq__(self, other:DNASeqRecordWithCoordinates) -> bool:
        """ equal to """
        if self.seq != other.seq:
            return False
        if len(self.locations) != len(other.locations):
            return False
        for i, j in zip(self.locations, other.locations):
            if i != j:
                return False
        return True

    def __ne__(self, other:DNASeqRecordWithCoordinates) -> bool:
        """ not equal to """
        return not self == other

    def __gt__(self, other:DNASeqRecordWithCoordinates) -> bool:
        """ greater than """
        for i, j in zip(self.locations, other.locations):
            if i > j:
                return True
            if i < j:
                return False
        if len(self.locations) > len(other.locations):
            return True
        return False

    def __ge__(self, other:DNASeqRecordWithCoordinates) -> bool:
        """ greater or equal to """
        return self > other or self == other

    def __lt__(self, other:DNASeqRecordWithCoordinates) -> bool:
        """ less than """
        return not self >= other

    def __le__(self, other:DNASeqRecordWithCoordinates) -> bool:
        """ less or equal """
        return not self > other

    def __le___(self, other:DNASeqRecordWithCoordinates):
        """ Due to an typo in biopython """
        return not self >= other

    def concat(self, other:DNASeqRecordWithCoordinates, _id:str=None,
            name:str=None, description:str=None
            ) -> DNASeqRecordWithCoordinates:
        """ concat two sequences into one """
        return self.__class__(
            seq=self.seq + other.seq,
            locations=self.locations + other.locations,
            orf=self.orf,
            id=self.id if _id is None else _id,
            name=self.name if name is None else name,
            description=self.description if description is None else \
                description
        )

    def search_orf(self):
        """ Searches the the open reading frame, finds the start and end
        position at the reference sequence and replace the `self.orf`. If
        no orf is found, (-1, -1) is returned. """
        start = self.find_orf_first()
        if start == -1:
            end = -1
        else:
            end = self.find_stop_codon(start)
            if end == -1:
                end = len(self) - len(self) % 3
        self.orf = FeatureLocation(
            start=start,
            end=end
        )

    def get_query_index(self, ref_index:int) -> int:
        """ Returns the query index wiht a given reference index """
        for location in self.locations:
            if ref_index in location.ref:
                return location.query.start + ref_index - location.ref.start
        return -1



""" Module for Gene Annotation Model """
from typing import List
from moPepGen.SeqFeature import FeatureLocation, MatchedLocation
from moPepGen import dna
from moPepGen.gtf.GTFSeqFeature import GTFSeqFeature


class GeneAnnotationModel(GTFSeqFeature):
    """ A GeneAnnotationModel object holds the genomic range of a gene, and
    the transcript IDs that it is associated with. """
    def __init__(self, chrom:str, attributes:dict, *args,
            transcripts:List[str]=None, **kwargs):
        """ Constructor """
        super().__init__(chrom=chrom, attributes=attributes, *args, **kwargs)
        if transcripts is None:
            transcripts = []
        self.transcripts = transcripts
        self.exons:List[GTFSeqFeature] = []

    def get_gene_sequence(self, chrom:dna.DNASeqRecord
            ) -> dna.DNASeqRecordWithCoordinates:
        """ Returns the DNA sequence of the gene, with all exons and introns
        reserved. """
        if self.strand == 1:
            seq = chrom.seq[self.location.start:self.location.end]
        elif self.strand == -1:
            seq = chrom.seq[self.location.start:self.location.end]
            seq = seq.reverse_complement()
        else:
            raise ValueError('Gene is unstranded.')
        gene_id = self.gene_id
        location = FeatureLocation(seqname=gene_id, start=0, end=len(seq))
        location = MatchedLocation(query=location, ref=location)
        return dna.DNASeqRecordWithCoordinates(seq=seq, locations=[location])

    def is_mrna_start_nf(self) -> bool:
        """ Whether the gene annotation has the mrna_start_NF tag """
        return 'tag' in self.attributes and \
            'mrna_start_NF' in self.attributes['tag']

    def is_cds_start_nf(self) -> bool:
        """ Whether the gene annotation has the cds_start_NF tag """
        return 'tag' in self.attributes and \
            'cds_start_NF' in self.attributes['tag']



""" `parseREDItools` takes RNA editing results called by
[REDItools](https://github.com/BioinfoUNIBA/REDItools) and saves them as a GVF
file. The GVF file can then be used to call variant peptides using
[callVariant](call-variant.md)
"""
from __future__ import annotations
import argparse
from pathlib import Path
from typing import Dict, List
from moPepGen import get_logger, seqvar, parser
from moPepGen.cli import common


INPUT_FILE_FORMATS = ['.tsv', '.txt']
OUTPUT_FILE_FORMATS = ['.gvf']

# pylint: disable=W0212
def add_subparser_parse_reditools(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen parseREDItools """

    p:argparse.ArgumentParser = subparsers.add_parser(
        name='parseREDItools',
        help='Parse REDItools result for moPepGen to call variant peptides.',
        description='Parse the REDItools result to a GVF format of variant'
        ' records for moPepGen to call variant peptides. The genome',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=p, formats=INPUT_FILE_FORMATS,
        message="File path to REDItools' TSV output."
    )
    common.add_args_output_path(p, OUTPUT_FILE_FORMATS)
    p.add_argument(
        '--transcript-id-column',
        type=int,
        help='The column index for transcript ID. If your REDItools table does'
        ' not contains it, use the AnnotateTable.py from the REDItools'
        ' package.',
        default=17,
        metavar='<number>'
    )
    p.add_argument(
        '--min-coverage-alt',
        type=int,
        help='Minimal read coverage of alterations to be parsed.',
        default=3,
        metavar='<number>'
    )
    p.add_argument(
        '--min-frequency-alt',
        type=float,
        help='Minimal frequency of alteration to be parsed.',
        default=0.1,
        metavar='<value>'
    )
    p.add_argument(
        '--min-coverage-rna',
        type=int,
        help='Minimal read coverage at the alteration site of RNAseq data of'
        ' reference and all alterations.',
        default=10,
        metavar='<value>'
    )
    p.add_argument(
        '--min-coverage-dna',
        type=int,
        help='Minimal read coverage at the alteration site of WGS. Set it to'
        ' -1 to skip checking this.',
        default=10,
        metavar='<number>'
    )
    common.add_args_source(p)
    common.add_args_reference(p, genome=False, proteome=False)
    common.add_args_debug_level(p)
    p.set_defaults(func=parse_reditools)
    common.print_help_if_missing_args(p)
    return p

def parse_reditools(args:argparse.Namespace) -> None:
    """ Parse REDItools output and save it in the GVF format. """
    logger = get_logger()
    # unpack args
    table_file:Path = args.input_path
    output_path:Path = args.output_path
    common.validate_file_format(
        table_file, INPUT_FILE_FORMATS, check_readable=True
    )
    common.validate_file_format(
        output_path, OUTPUT_FILE_FORMATS, check_writable=True
    )

    transcript_id_column = args.transcript_id_column - 1
    min_coverage_alt:int = args.min_coverage_alt
    min_frequency_alt:int = args.min_frequency_alt
    min_coverage_rna:int = args.min_coverage_rna
    min_coverage_dna:int = args.min_coverage_dna

    common.print_start_message(args)

    _, anno, *_ = common.load_references(args, load_genome=False, load_canonical_peptides=False)

    variants:Dict[str, List[seqvar.VariantRecord]] = {}

    for record in parser.REDItoolsParser.parse(table_file, transcript_id_column):
        _vars = record.convert_to_variant_records(
            anno=anno,
            min_coverage_alt=min_coverage_alt,
            min_frequency_alt=min_frequency_alt,
            min_coverage_rna=min_coverage_rna,
            min_coverage_dna=min_coverage_dna
        )
        for variant in _vars:
            gene_id = variant.location.seqname
            if gene_id not in variants:
                variants[gene_id] = []
            variants[gene_id].append(variant)

    logger.info('REDItools table %s loaded.', table_file)

    if not variants:
        logger.warning('No variant record is saved.')
        return

    for records in variants.values():
        records.sort()

    logger.info('Variants sorted.')

    metadata = common.generate_metadata(args)

    genes_rank = anno.get_genes_rank()
    ordered_keys = sorted(variants.keys(), key=lambda x:genes_rank[x])

    all_records = []
    for key in ordered_keys:
        records = variants[key]
        all_records.extend(records)

    seqvar.io.write(all_records, output_path, metadata)

    logger.info('Variants written to disk.')



""" Module for rMATS parser """
from __future__ import annotations
from typing import TYPE_CHECKING, List, Tuple
from moPepGen import gtf, dna, seqvar
from .RMATSRecord import RMATSRecord


if TYPE_CHECKING:
    from moPepGen.seqvar import SpliceJunction


class SERecord(RMATSRecord):
    """ Skipped Exon """
    def __init__(self, gene_id:str, gene_symbol:str, chrom:str, exon_start:int,
            exon_end:int, upstream_exon_start:int, upstream_exon_end:int,
            downstream_exon_start:int, downstream_exon_end:int,
            ijc_sample_1:int, sjc_sample_1:int, ijc_sample_2:int,
            sjc_sample_2:int, inc_form_len:int, skip_form_len:int,
            pvalue:float, fdr:float):
        """ Constructor """
        super().__init__(gene_id, gene_symbol, chrom)
        self.exon_start = exon_start
        self.exon_end = exon_end
        self.upstream_exon_start = upstream_exon_start
        self.upstream_exon_end = upstream_exon_end
        self.downstream_exon_start = downstream_exon_start
        self.downstream_exon_end = downstream_exon_end
        self.ijc_sample_1 = ijc_sample_1
        self.sjc_sample_1 = sjc_sample_1
        self.ijc_sample_2 = ijc_sample_2
        self.sjc_sample_2 = sjc_sample_2
        self.inc_form_len = inc_form_len
        self.skip_form_len = skip_form_len
        self.pvalue = pvalue
        self.fdr = fdr

    @classmethod
    def readline(cls, line:str) -> SERecord:
        """ Create a SERecord from a line of the rMATS output """
        fields = line.rstrip().split('\t')
        return cls(
            gene_id=fields[1].strip('"'),
            gene_symbol=fields[2].strip('"'),
            chrom=fields[3],
            exon_start=int(fields[5]),
            exon_end=int(fields[6]),
            upstream_exon_start=int(fields[7]),
            upstream_exon_end=int(fields[8]),
            downstream_exon_start=int(fields[9]),
            downstream_exon_end=int(fields[10]),
            ijc_sample_1=int(fields[12]),
            sjc_sample_1=int(fields[13]),
            ijc_sample_2=None if fields[14] == '' else int(fields[14]),
            sjc_sample_2=None if fields[15] == '' else int(fields[15]),
            inc_form_len=int(fields[16]),
            skip_form_len=int(fields[17]),
            pvalue=None if fields[18] == 'NA' else float(fields[18]),
            fdr=None if fields[19] == 'NA' else float(fields[19])
        )

    def create_variant_id(self, anno:gtf.GenomicAnnotation) -> str:
        """ Create variant ID """
        uee = anno.coordinate_genomic_to_gene(self.upstream_exon_end - 1, self.gene_id)
        es = anno.coordinate_genomic_to_gene(self.exon_start, self.gene_id)
        ee = anno.coordinate_genomic_to_gene(self.exon_end - 1, self.gene_id)
        des = anno.coordinate_genomic_to_gene(self.downstream_exon_start, self.gene_id)

        gene_model = anno.genes[self.gene_id]
        if gene_model.strand == -1:
            es, ee = ee, es
            uee, des = des, uee
        ee += 1
        des += 1

        return f"SE_{uee}-{es}-{ee}-{des}"

    def create_splice_junctions(self
            ) -> Tuple[SpliceJunction, SpliceJunction, SpliceJunction]:
        """ Create splice junctions """
        skip_junction = seqvar.SpliceJunction(
            upstream_start=self.upstream_exon_start,
            upstream_end=self.upstream_exon_end,
            downstream_start=self.downstream_exon_start,
            downstream_end=self.downstream_exon_end,
            gene_id=self.gene_id,
            chrom=self.chrom
        )
        upstream_junction = seqvar.SpliceJunction(
            upstream_start=self.upstream_exon_start,
            upstream_end=self.upstream_exon_end,
            downstream_start=self.exon_start,
            downstream_end=self.exon_end,
            gene_id=self.gene_id,
            chrom=self.chrom
        )
        downstream_junction = seqvar.SpliceJunction(
            upstream_start=self.exon_start,
            upstream_end=self.exon_end,
            downstream_start=self.downstream_exon_start,
            downstream_end=self.downstream_exon_end,
            gene_id=self.gene_id,
            chrom=self.chrom
        )
        return skip_junction, upstream_junction, downstream_junction

    def convert_to_variant_records(self, anno:gtf.GenomicAnnotation,
            genome:dna.DNASeqDict, min_ijc:int, min_sjc:int
            ) -> List[seqvar.VariantRecord]:
        """ Convert to variant records """
        variants = []

        gene_model = anno.genes[self.gene_id]

        skip_junction, upstream_junction, downstream_junction \
            = self.create_splice_junctions()

        if not skip_junction.is_novel(anno) \
                and not upstream_junction.is_novel(anno)\
                and not downstream_junction.is_novel(anno):
            return variants

        tx_ids = gene_model.transcripts
        chrom = gene_model.location.seqname
        gene_seq = gene_model.get_gene_sequence(genome[chrom])
        var_id = self.create_variant_id(anno)

        for tx_id in tx_ids:
            tx_model = anno.transcripts[tx_id]

            if self.sjc_sample_1 >= min_sjc:
                aln = skip_junction.align_to_transcript(tx_model, False, False)
                if aln:
                    variants += aln.convert_to_variant_records(anno, gene_seq, var_id)

            if self.ijc_sample_1 >= min_ijc:
                aln = upstream_junction.align_to_transcript( tx_model, False, True)
                if aln:
                    variants += aln.convert_to_variant_records(anno, gene_seq, var_id)
                aln = downstream_junction.align_to_transcript(tx_model, True, False)
                if aln:
                    variants += aln.convert_to_variant_records(anno, gene_seq, var_id)

        return variants
